# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for Estimator."""

import unittest
from test import QiskitTestCase

import numpy as np

from qiskit.circuit import Parameter, QuantumCircuit
from qiskit.circuit.library import real_amplitudes
from qiskit.primitives import StatevectorEstimator
from qiskit.primitives.containers.bindings_array import BindingsArray
from qiskit.primitives.containers.estimator_pub import EstimatorPub
from qiskit.primitives.containers.observables_array import ObservablesArray
from qiskit.quantum_info import SparsePauliOp


class TestStatevectorEstimator(QiskitTestCase):
    """Test Estimator"""

    def setUp(self):
        super().setUp()
        self.ansatz = QuantumCircuit(2)
        self.ansatz.append(real_amplitudes(num_qubits=2, reps=2), [0, 1])

        self.observable = SparsePauliOp.from_list(
            [
                ("II", -1.052373245772859),
                ("IZ", 0.39793742484318045),
                ("ZI", -0.39793742484318045),
                ("ZZ", -0.01128010425623538),
                ("XX", 0.18093119978423156),
            ]
        )
        self.expvals = -1.0284380963435145, -1.284366511861733

        ra_2_reps = QuantumCircuit(2)
        ra_2_reps.append(real_amplitudes(num_qubits=2, reps=2), [0, 1])
        ra_3_reps = QuantumCircuit(2)
        ra_3_reps.append(real_amplitudes(num_qubits=2, reps=3), [0, 1])
        self.psi = (ra_2_reps, ra_3_reps)

        self.params = tuple(psi.parameters for psi in self.psi)
        self.hamiltonian = (
            SparsePauliOp.from_list([("II", 1), ("IZ", 2), ("XI", 3)]),
            SparsePauliOp.from_list([("IZ", 1)]),
            SparsePauliOp.from_list([("ZI", 1), ("ZZ", 1)]),
        )
        self.theta = (
            [0, 1, 1, 2, 3, 5],
            [0, 1, 1, 2, 3, 5, 8, 13],
            [1, 2, 3, 4, 5, 6],
        )

    def test_estimator_run(self):
        """Test Estimator.run()"""
        psi1, psi2 = self.psi
        hamiltonian1, hamiltonian2, hamiltonian3 = self.hamiltonian
        theta1, theta2, theta3 = self.theta
        estimator = StatevectorEstimator()

        # Specify the circuit and observable by indices.
        # calculate [ <psi1(theta1)|H1|psi1(theta1)> ]
        job = estimator.run([(psi1, hamiltonian1, [theta1])])
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.5555572817900956])

        # Objects can be passed instead of indices.
        # Note that passing objects has an overhead
        # since the corresponding indices need to be searched.
        # User can append a circuit and observable.
        # calculate [ <psi2(theta2)|H1|psi2(theta2)> ]
        result2 = estimator.run([(psi2, hamiltonian1, theta2)]).result()
        np.testing.assert_allclose(result2[0].data.evs, [2.97797666])

        # calculate [ <psi1(theta1)|H2|psi1(theta1)>, <psi1(theta1)|H3|psi1(theta1)> ]
        result3 = estimator.run([(psi1, [hamiltonian2, hamiltonian3], theta1)]).result()
        np.testing.assert_allclose(result3[0].data.evs, [-0.551653, 0.07535239])

        # calculate [ [<psi1(theta1)|H1|psi1(theta1)>,
        #              <psi1(theta3)|H3|psi1(theta3)>],
        #             [<psi2(theta2)|H2|psi2(theta2)>] ]
        result4 = estimator.run(
            [(psi1, [hamiltonian1, hamiltonian3], [theta1, theta3]), (psi2, hamiltonian2, theta2)]
        ).result()
        np.testing.assert_allclose(result4[0].data.evs, [1.55555728, -1.08766318])
        np.testing.assert_allclose(result4[1].data.evs, [0.17849238])

    def test_estimator_with_pub(self):
        """Test estimator with explicit EstimatorPubs."""
        psi1, psi2 = self.psi
        hamiltonian1, hamiltonian2, hamiltonian3 = self.hamiltonian
        theta1, theta2, theta3 = self.theta

        obs1 = ObservablesArray.coerce([hamiltonian1, hamiltonian3])
        bind1 = BindingsArray.coerce({tuple(psi1.parameters): [theta1, theta3]})
        pub1 = EstimatorPub(psi1, obs1, bind1)
        obs2 = ObservablesArray.coerce(hamiltonian2)
        bind2 = BindingsArray.coerce({tuple(psi2.parameters): theta2})
        pub2 = EstimatorPub(psi2, obs2, bind2)

        estimator = StatevectorEstimator()
        result4 = estimator.run([pub1, pub2]).result()
        np.testing.assert_allclose(result4[0].data.evs, [1.55555728, -1.08766318])
        np.testing.assert_allclose(result4[1].data.evs, [0.17849238])

    def test_estimator_run_no_params(self):
        """test for estimator without parameters"""
        circuit = self.ansatz.assign_parameters([0, 1, 1, 2, 3, 5])
        est = StatevectorEstimator()
        result = est.run([(circuit, self.observable)]).result()
        np.testing.assert_allclose(result[0].data.evs, [-1.284366511861733])

    def test_run_single_circuit_observable(self):
        """Test for single circuit and single observable case."""
        est = StatevectorEstimator()

        with self.subTest("No parameter"):
            qc = QuantumCircuit(1)
            qc.x(0)
            op = SparsePauliOp("Z")
            param_vals = [None, [], [[]], np.array([]), np.array([[]]), [np.array([])]]
            target = [-1]
            for val in param_vals:
                self.subTest(f"{val}")
                result = est.run([(qc, op, val)]).result()
                np.testing.assert_allclose(result[0].data.evs, target)
                self.assertEqual(result[0].metadata["target_precision"], 0)

        with self.subTest("One parameter"):
            param = Parameter("x")
            qc = QuantumCircuit(1)
            qc.ry(param, 0)
            op = SparsePauliOp("Z")
            param_vals = [
                [np.pi],
                np.array([np.pi]),
            ]
            target = [-1]
            for val in param_vals:
                self.subTest(f"{val}")
                result = est.run([(qc, op, val)]).result()
                np.testing.assert_allclose(result[0].data.evs, target)
                self.assertEqual(result[0].metadata["target_precision"], 0)

        with self.subTest("More than one parameter"):
            qc = self.psi[0]
            op = self.hamiltonian[0]
            param_vals = [
                self.theta[0],
                [self.theta[0]],
                np.array(self.theta[0]),
                np.array([self.theta[0]]),
                [np.array(self.theta[0])],
            ]
            target = [1.5555572817900956]
            for val in param_vals:
                self.subTest(f"{val}")
                result = est.run([(qc, op, val)]).result()
                np.testing.assert_allclose(result[0].data.evs, target)
                self.assertEqual(result[0].metadata["target_precision"], 0)

    def test_run_1qubit(self):
        """Test for 1-qubit cases"""
        qc = QuantumCircuit(1)
        qc2 = QuantumCircuit(1)
        qc2.x(0)

        op = SparsePauliOp.from_list([("I", 1)])
        op2 = SparsePauliOp.from_list([("Z", 1)])

        est = StatevectorEstimator()
        result = est.run([(qc, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [-1])

    def test_run_2qubits(self):
        """Test for 2-qubit cases (to check endian)"""
        qc = QuantumCircuit(2)
        qc2 = QuantumCircuit(2)
        qc2.x(0)

        op = SparsePauliOp.from_list([("II", 1)])
        op2 = SparsePauliOp.from_list([("ZI", 1)])
        op3 = SparsePauliOp.from_list([("IZ", 1)])

        est = StatevectorEstimator()
        result = est.run([(qc, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc, op3)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op3)]).result()
        np.testing.assert_allclose(result[0].data.evs, [-1])

    def test_run_errors(self):
        """Test for errors"""
        qc = QuantumCircuit(1)
        qc2 = QuantumCircuit(2)

        op = SparsePauliOp.from_list([("I", 1)])
        op2 = SparsePauliOp.from_list([("II", 1)])

        est = StatevectorEstimator()
        with self.assertRaises(ValueError):
            est.run([(qc, op2)]).result()
        with self.assertRaises(ValueError):
            est.run([(qc, op, [[1e4]])]).result()
        with self.assertRaises(ValueError):
            est.run([(qc2, op2, [[1, 2]])]).result()
        with self.assertRaises(ValueError):
            est.run([(qc, [op, op2], [[1]])]).result()
        with self.assertRaises(ValueError):
            est.run([(qc, op)], precision=-1).result()
        with self.assertRaises(ValueError):
            est.run([(qc, 1j * op)], precision=0.1).result()
        with self.subTest("missing []"):
            with self.assertRaisesRegex(ValueError, "An invalid Estimator pub-like was given"):
                _ = est.run((qc, op)).result()

    def test_run_numpy_params(self):
        """Test for numpy array as parameter values"""
        qc = QuantumCircuit(2)
        qc.append(real_amplitudes(num_qubits=2, reps=2), [0, 1])
        op = SparsePauliOp.from_list([("IZ", 1), ("XI", 2), ("ZY", -1)])
        k = 5
        rng = np.random.default_rng(12)
        params_array = rng.random((k, qc.num_parameters))
        params_list = params_array.tolist()
        params_list_array = list(params_array)
        estimator = StatevectorEstimator()
        target = estimator.run([(qc, op, params_list)]).result()

        with self.subTest("ndarrary"):
            result = estimator.run([(qc, op, params_array)]).result()
            self.assertEqual(len(result[0].data.evs), k)
            np.testing.assert_allclose(result[0].data.evs, target[0].data.evs)

        with self.subTest("list of ndarray"):
            result = estimator.run([(qc, op, params_list_array)]).result()
            self.assertEqual(len(result[0].data.evs), k)
            np.testing.assert_allclose(result[0].data.evs, target[0].data.evs)

    def test_precision_seed(self):
        """Test for precision and seed"""
        estimator = StatevectorEstimator(default_precision=1.0, seed=1)
        psi1 = self.psi[0]
        hamiltonian1 = self.hamiltonian[0]
        theta1 = self.theta[0]
        job = estimator.run([(psi1, hamiltonian1, [theta1])])
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.901141473854881])
        # The result of the second run is the same
        job = estimator.run([(psi1, hamiltonian1, [theta1]), (psi1, hamiltonian1, [theta1])])
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.901141473854881])
        np.testing.assert_allclose(result[1].data.evs, [1.901141473854881])
        # precision=0 implies the exact expectation value
        job = estimator.run([(psi1, hamiltonian1, [theta1])], precision=0)
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.5555572817900956])

    def test_iter_pub(self):
        """test for an iterable of pubs"""
        estimator = StatevectorEstimator()
        circuit = self.ansatz.assign_parameters([0, 1, 1, 2, 3, 5])
        observable = self.observable.apply_layout(circuit.layout)
        result = estimator.run(iter([(circuit, observable), (circuit, observable)])).result()
        np.testing.assert_allclose(result[0].data.evs, [-1.284366511861733])
        np.testing.assert_allclose(result[1].data.evs, [-1.284366511861733])

    def test_metadata(self):
        """Test for metadata"""
        qc = QuantumCircuit(2)
        qc2 = QuantumCircuit(2)
        qc2.metadata = {"a": 1}
        estimator = StatevectorEstimator()
        result = estimator.run([(qc, "ZZ"), (qc2, "ZZ")], precision=0.1).result()

        self.assertEqual(len(result), 2)
        self.assertEqual(result.metadata, {"version": 2})
        self.assertEqual(
            result[0].metadata, {"target_precision": 0.1, "circuit_metadata": qc.metadata}
        )
        self.assertEqual(
            result[1].metadata, {"target_precision": 0.1, "circuit_metadata": qc2.metadata}
        )

    def test_reset(self):
        """Test for circuits with reset."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        qc.reset(0)
        op = SparsePauliOp("ZI")

        seed = 12
        n = 1000
        estimator = StatevectorEstimator(seed=seed)
        with self.subTest("precision=0"):
            result = estimator.run([(qc, [op] * n)]).result()
            # expectation values should be stochastic due to reset for subsystems
            np.testing.assert_allclose(result[0].data.evs.mean(), 0, atol=1e-1)

            result2 = estimator.run([(qc, [op] * n)]).result()
            # expectation values should be reproducible due to seed
            np.testing.assert_allclose(result[0].data.evs, result2[0].data.evs)

        with self.subTest("precision=0.01"):
            precision = 0.01
            result = estimator.run([(qc, [op] * n)], precision=precision).result()
            # expectation values should be stochastic due to reset for subsystems
            np.testing.assert_allclose(result[0].data.evs.mean(), 0, atol=1e-1)

            result2 = estimator.run([(qc, [op] * n)], precision=precision).result()
            # expectation values should be reproducible due to seed
            np.testing.assert_allclose(result[0].data.evs, result2[0].data.evs)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Functions to generate the basic approximations of single qubit gates for Solovay-Kitaev."""

from __future__ import annotations

from qiskit.circuit import Gate
from qiskit.utils.deprecation import deprecate_func
from qiskit._accelerate.synthesis.discrete_basis import GateSequence

from .solovay_kitaev import SolovayKitaevDecomposition


@deprecate_func(
    since="2.1",
    additional_msg=(
        "Use the SolovayKitaevDecomposition class directly, to generate, store, and load the "
        "basic approximations."
    ),
    pending=True,
)
def generate_basic_approximations(
    basis_gates: list[str | Gate], depth: int, filename: str | None = None
) -> list[GateSequence]:
    """Generates a list of :class:`GateSequence`\\ s with the gates in ``basis_gates``.

    Args:
        basis_gates: The gates from which to create the sequences of gates.
        depth: The maximum depth of the approximations.
        filename: If provided, the basic approximations are stored in this file.

    Returns:
        List of :class:`GateSequence`\\ s using the gates in ``basis_gates``.

    Raises:
        ValueError: If ``basis_gates`` contains an invalid gate identifier.
    """
    sk = SolovayKitaevDecomposition(basis_gates=basis_gates, depth=depth)

    if filename is not None:
        sk.save_basic_approximations(filename)

    return sk._sk.get_gate_sequences()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test parameter expression."""

import cmath
import math
import unittest

from test import combine
from test import QiskitTestCase

import ddt

from qiskit.circuit import Parameter, ParameterVector, ParameterExpression
from qiskit.utils.optionals import HAS_SYMPY


param_x = Parameter("x")
param_y = Parameter("y")
nested_expr = param_x + param_y - param_x
nested_expr = nested_expr.subs({param_y: param_x})

vector = ParameterVector("vec", 1000)
nested_vector_expr = vector[500] + vector[256] - vector[500]
for i in range(1000):
    nested_vector_expr += vector[i] - vector[i]


operands = [
    Parameter("a"),
    Parameter("dai"),
    ParameterVector("a", 100)[42],
    complex(3.14, -3.14),
    complex(1.0, 1.0),
    complex(0, 1),
    complex(-1, 0),
    2.3,
    int(5),
    int(-5),
    1.0,
    -1.0,
    0,
    0.0,
    complex(0, 0),
    Parameter("ab") + 2 - 2,
    Parameter("abc") ** 1.0,
    Parameter("abcd") / 1,
    Parameter("X") * 1.0,
    Parameter("Y") ** complex(1.0, 0),
    Parameter("abcd_complex") / complex(1, 0),
    ParameterVector("b", 1)[0] + (0 * 1) * Parameter("ZERO"),
    nested_expr,
    nested_vector_expr,
]

bind_values = [math.pi, -math.pi, 5, -5, complex(2, 1), complex(-1, 2), 0, complex(0, 0)]
real_values = [0.41, 0.9, -0.83, math.pi, -math.pi / 124, -42.42]


@ddt.ddt
class TestParameterExpression(QiskitTestCase):
    """Test parameter expression."""

    @combine(
        left=operands,
        right=operands,
        bind_value=bind_values,
        name="{left}_plus_{right}_bind_{bind_value}",
    )
    def test_addition_simple(self, left, right, bind_value):
        """Test expression addition."""
        if isinstance(left, ParameterExpression) or isinstance(right, ParameterExpression):
            expr = left + right
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(left, ParameterExpression) and isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), 2.0 * bind_value)
            elif not isinstance(left, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), left + bind_value)
            elif not isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), right + bind_value)

    @combine(
        left=operands,
        right=operands,
        bind_value=bind_values,
        name="{left}_minus_{right}_bind_{bind_value}",
    )
    def test_subtraction_simple(self, left, right, bind_value):
        """Test expression subtraction."""
        if isinstance(left, ParameterExpression) or isinstance(right, ParameterExpression):
            expr = left - right
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(left, ParameterExpression) and isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), 0.0)
            elif not isinstance(left, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), left - bind_value)
            elif not isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), bind_value - right)

    @combine(
        left=operands,
        right=operands,
        bind_value=bind_values,
        name="{left}_mul_{right}_bind_{bind_value}",
    )
    def test_multiplication_simple(self, left, right, bind_value):
        """Test expression multiplication."""
        if isinstance(left, ParameterExpression) or isinstance(right, ParameterExpression):
            expr = left * right
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(left, ParameterExpression) and isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), bind_value * bind_value)
            elif not isinstance(left, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), left * bind_value)
            elif not isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), bind_value * right)

    @combine(
        left=operands,
        right=operands,
        bind_value=bind_values,
        name="{left}_div_{right}_bind_{bind_value}",
    )
    def test_division_simple(self, left, right, bind_value):
        """Test expression division."""
        if isinstance(left, ParameterExpression) or isinstance(right, ParameterExpression):
            if not isinstance(right, ParameterExpression) and right == 0:
                with self.assertRaises(ZeroDivisionError):
                    _ = left / right
                return
            expr = left / right
            try:
                res = expr.bind({x: bind_value for x in expr.parameters})
            except ZeroDivisionError:
                self.assertIsInstance(right, ParameterExpression)
                self.assertAlmostEqual(bind_value, 0)
                return
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(left, ParameterExpression) and isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), 1.0)
            elif not isinstance(left, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), left / bind_value)
            elif not isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), bind_value / right)

    @combine(
        left=operands,
        right=operands,
        name="{left}_pow_{right}",
    )
    def test_pow_simple(self, left, right):
        """Test expression pow."""
        if isinstance(left, ParameterExpression) or isinstance(right, ParameterExpression):
            expr = left**right
            res = expr.bind({x: 1.0 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(left, ParameterExpression) and isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), 1.0)
            elif not isinstance(left, ParameterExpression):
                if isinstance(left, complex):
                    self.assertAlmostEqual(res.numeric(), left)
                else:
                    self.assertAlmostEqual(res.numeric(), left)
            elif not isinstance(right, ParameterExpression):
                self.assertAlmostEqual(res.numeric(), 1.0**right)

    @combine(
        left=operands,
        right=operands,
        name="{left}_pow_{right}",
    )
    def test_pow_complex_binding(self, left, right):
        """Test expression pow with complex binding."""
        if isinstance(left, ParameterExpression) or isinstance(right, ParameterExpression):
            expr = left**right
            res = expr.bind({x: complex(1.0, 1.0) for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(left, ParameterExpression) and isinstance(right, ParameterExpression):
                self.assertAlmostEqual(complex(res), complex(1.0, 1.0) ** complex(1.0, 1.0))
            elif not isinstance(left, ParameterExpression):
                if left != 0:
                    self.assertAlmostEqual(complex(res), left ** complex(1.0, 1.0))
                else:
                    with self.assertRaises(ZeroDivisionError):
                        _ = left ** complex(1.0, 1.0)
            elif not isinstance(right, ParameterExpression):
                self.assertAlmostEqual(complex(res), complex(1.0, 1.0) ** right)

    def test_pow_creates_complex(self):
        """Test a complex is created when appropriate."""
        param_a = Parameter("A")
        param_b = Parameter("B")
        param_c = Parameter("C")
        expr = param_a + param_b + param_c
        expr = expr.subs({param_b: param_a + 2 * param_c})
        expr = expr.subs({param_a: -param_a, param_c: -param_c})
        expr = expr**0.5
        res = expr.bind({param_a: 2, param_c: 2})
        self.assertFalse(res.is_real())
        # Expected is sqrt(-10):
        self.assertAlmostEqual(complex(0, 3.1622776601683795), complex(res))

    @combine(expression=operands, bind_value=bind_values, name="{expression}_bind_{bind_value}")
    def test_abs_simple(self, expression, bind_value):
        """Test expression abs."""
        if isinstance(expression, ParameterExpression):
            expr = abs(expression)
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, abs(bind_value))

    @combine(expression=operands)
    def test_acos_simple(self, expression):
        """Test expression arccos."""
        if isinstance(expression, ParameterExpression):
            expr = expression.arccos()
            res = expr.bind({x: 0.2 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, math.acos(0.2))
            # Test negative
            expr = expression.arccos()
            res = expr.bind({x: -0.3 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, math.acos(-0.3))

    @combine(expression=operands)
    def test_asin_simple(self, expression):
        """Test expression arcsin."""
        if isinstance(expression, ParameterExpression):
            expr = expression.arcsin()
            res = expr.bind({x: 0.2 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, math.asin(0.2))
            # Test negative
            expr = expression.arcsin()
            res = expr.bind({x: -0.3 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, math.asin(-0.3))

    @combine(expression=operands)
    def test_atan_simple(self, expression):
        """Test expression arctan."""
        if isinstance(expression, ParameterExpression):
            expr = expression.arctan()
            res = expr.bind({x: 0.2 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, math.atan(0.2))
            # Test negative
            expr = expression.arctan()
            res = expr.bind({x: -0.3 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, math.atan(-0.3))

    @combine(expression=operands, bind_value=bind_values, name="{expression}_bind_{bind_value}")
    def test_conjugate_simple(self, expression, bind_value):
        """Test expression conjugate."""
        if isinstance(expression, ParameterExpression):
            expr = expression.conjugate()
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            numeric = res.numeric()
            if isinstance(numeric, complex):
                self.assertEqual(res, bind_value.conjugate())
            else:
                self.assertEqual(res, bind_value)

    @combine(expression=operands, bind_value=bind_values, name="{expression}_bind_{bind_value}")
    def test_cos_simple(self, expression, bind_value):
        """Test expression cos."""
        if isinstance(expression, ParameterExpression):
            expr = expression.cos()
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(bind_value, complex):
                self.assertAlmostEqual(res.numeric(), cmath.cos(bind_value))
            else:
                self.assertAlmostEqual(res.numeric(), math.cos(bind_value))

    @combine(expression=operands, bind_value=bind_values, name="{expression}_bind_{bind_value}")
    def test_sin_simple(self, expression, bind_value):
        """Test expression sin."""
        if isinstance(expression, ParameterExpression):
            expr = expression.sin()
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(bind_value, complex):
                self.assertAlmostEqual(res.numeric(), cmath.sin(bind_value))
            else:
                self.assertAlmostEqual(res.numeric(), math.sin(bind_value))

    @combine(expression=operands, bind_value=bind_values, name="{expression}_bind_{bind_value}")
    def test_tan_simple(self, expression, bind_value):
        """Test expression tan."""
        if isinstance(expression, ParameterExpression):
            expr = expression.tan()
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(bind_value, complex):
                self.assertAlmostEqual(res.numeric(), cmath.tan(bind_value))
            else:
                self.assertAlmostEqual(res.numeric(), math.tan(bind_value))

    @combine(expression=operands, bind_value=bind_values, name="{expression}_bind_{bind_value}")
    def test_exp_simple(self, expression, bind_value):
        """Test expression exp."""
        if isinstance(expression, ParameterExpression):
            expr = expression.exp()
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(bind_value, complex):
                self.assertAlmostEqual(res.numeric(), cmath.exp(bind_value))
            else:
                self.assertAlmostEqual(res.numeric(), math.exp(bind_value))

    @combine(expression=operands, bind_value=bind_values, name="{expression}_bind_{bind_value}")
    def test_log_simple(self, expression, bind_value):
        """Test expression log."""
        if isinstance(expression, ParameterExpression) and bind_value != 0:
            expr = expression.log()
            res = expr.bind({x: bind_value for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            if isinstance(bind_value, complex):
                self.assertAlmostEqual(res.numeric(), cmath.log(bind_value))
            else:
                if bind_value > 0:
                    self.assertAlmostEqual(res.numeric(), math.log(bind_value))
                else:
                    self.assertAlmostEqual(res.numeric(), cmath.log(bind_value))

    @combine(expression=operands)
    def test_sign_simple(self, expression):
        """Test expression sign."""
        if isinstance(expression, ParameterExpression) and expression.is_real():
            expr = expression.sign()
            res = expr.bind({x: -0.1 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, -1)
            expr = expression.sign()
            res = expr.bind({x: 0.1 for x in expr.parameters})
            self.assertEqual(res, 1)
            expr = expression.sign()
            res = expr.bind({x: 0.0 for x in expr.parameters})
            self.assertEqual(res, 0)
            expr = expression.sign()
            res = expr.bind({x: -2 for x in expr.parameters})
            self.assertIsInstance(res, ParameterExpression)
            self.assertEqual(res, -1)
            expr = expression.sign()
            res = expr.bind({x: 5 for x in expr.parameters})
            self.assertEqual(res, 1)
            expr = expression.sign()
            res = expr.bind({x: 0 for x in expr.parameters})
            self.assertEqual(res, 0)

    @combine(expression=operands)
    def test_is_real(self, expression):
        """Test the is_real() method."""
        if isinstance(expression, ParameterExpression):
            res = expression.bind({x: complex(1.0, 0.0) for x in expression.parameters})
            self.assertTrue(res.is_real())
            res = expression.bind({x: complex(1.0, 1.0) for x in expression.parameters})
            self.assertFalse(res.is_real())
            res = expression.bind({x: 1.0 for x in expression.parameters})
            self.assertTrue(res.is_real())
            res = expression.bind({x: 5 for x in expression.parameters})
            self.assertTrue(res.is_real())
            self.assertFalse(expression.is_real())

    @combine(expression=operands)
    def test_casting(self, expression):
        """Test casting"""
        if isinstance(expression, ParameterExpression):
            res = expression.bind({x: complex(1.0, 0.0) for x in expression.parameters})
            self.assertIsInstance(complex(res), complex)
            self.assertIsInstance(float(res), float)
            self.assertIsInstance(int(res), int)
            self.assertEqual(res, 1)
            res = expression.bind({x: complex(1.0, 1.0) for x in expression.parameters})
            self.assertIsInstance(complex(res), complex)
            with self.assertRaises(TypeError):
                float(res)
            with self.assertRaises(TypeError):
                int(res)
            self.assertAlmostEqual(complex(res), complex(1.0, 1.0))
            res = expression.bind({x: 1.0 for x in expression.parameters})
            self.assertIsInstance(complex(res), complex)
            self.assertIsInstance(float(res), float)
            self.assertIsInstance(int(res), int)
            self.assertEqual(res, 1.0)
            res = expression.bind({x: 5 for x in expression.parameters})
            self.assertIsInstance(complex(res), complex)
            self.assertIsInstance(float(res), float)
            self.assertIsInstance(int(res), int)
            self.assertEqual(res, 5)

    @combine(expression=operands)
    def test_numeric(self, expression):
        """Test numeric"""
        if isinstance(expression, ParameterExpression):
            res = expression.bind({x: complex(1.0, 0.0) for x in expression.parameters}).numeric()
            self.assertIsInstance(res, float)
            self.assertEqual(res, 1)
            res = expression.bind({x: complex(1.0, 1.0) for x in expression.parameters}).numeric()
            self.assertIsInstance(res, complex)
            self.assertAlmostEqual(complex(res), complex(1.0, 1.0))
            res = expression.bind({x: 1.0 for x in expression.parameters}).numeric()
            self.assertIsInstance(float(res), float)
            self.assertEqual(res, 1.0)
            res = expression.bind({x: 5 for x in expression.parameters}).numeric()
            self.assertIsInstance(int(res), int)
            self.assertEqual(res, 5)

    def test_derivatives_numeric(self):
        """Test derivatives with numerical values."""
        methods_and_references = [
            ("abs", lambda x: x / abs(x)),
            ("exp", math.exp),
            ("log", lambda x: 1 / x),
            ("sin", math.cos),
            ("cos", lambda x: -math.sin(x)),
            ("tan", lambda x: 1 / math.cos(x) ** 2),
            ("arccos", lambda x: -((1 - x**2) ** (-0.5))),
            ("arcsin", lambda x: (1 - x**2) ** (-0.5)),
            ("arctan", lambda x: 1 / (1 + x**2)),
            ("conjugate", lambda x: 1),
        ]

        x = Parameter("x")
        for method, reference in methods_and_references:
            expr = getattr(x, method)()
            d_expr = expr.gradient(x)

            for value in real_values:
                if method == "log" and value <= 0:
                    continue  # log not defined
                if method in ["arccos", "arcsin", "arctan"] and abs(value) >= 1 - 1e-10:
                    continue  # arc-funcs not defined

                with self.subTest(method=method, value=value):
                    ref = reference(value)
                    if isinstance(d_expr, ParameterExpression):
                        val = d_expr.bind({x: value}).numeric()
                    else:
                        val = d_expr  # d/dx conj(x) == 1

                    self.assertAlmostEqual(ref, val)

    def test_sign_derivative_errors(self):
        """Test the derivative of sign errors (not supported right now)."""
        x = Parameter("x")
        expr = x.sign()

        with self.assertRaises(RuntimeError):
            _ = expr.gradient(x)

    @unittest.skipUnless(HAS_SYMPY, "Sympy is required for this test")
    def test_sympify_all_ops(self):
        """Test the sympify function works for all the supported operations."""

        import sympy

        a = Parameter("a")
        b = Parameter("b")
        c = Parameter("c")
        d = Parameter("d")

        expression = (a + b.sin() / 4) * c**2
        final_expr = (
            (expression.cos() + d.arccos() - d.arcsin() + d.arctan() + d.tan()) / d.exp()
            + expression.gradient(a)
            + expression.log().sign()
            - a.sin()
            - b.conjugate()
        )
        final_expr = final_expr.abs()
        final_expr = final_expr.subs({c: a})
        result = final_expr.sympify()

        a = sympy.Symbol("a")
        b = sympy.Symbol("b")
        c = sympy.Symbol("c")
        d = sympy.Symbol("d")
        expression = (a + sympy.sin(b) / 4) * c**2
        expected = (
            (sympy.cos(expression) + sympy.acos(d) - sympy.asin(d) + sympy.atan(d) + sympy.tan(d))
            / sympy.exp(d)
            + expression.diff(a)
            + sympy.sign(sympy.log(expression))
            - sympy.sin(a)
            - sympy.conjugate(b)
        )
        expected = sympy.Abs(expected)
        expected = expected.subs({c: a})
        self.assertEqual(result, expected)

    @unittest.skipUnless(HAS_SYMPY, "Sympy is required for this test")
    def test_sympify_subs_vector(self):
        """Test an expression with subbed ParameterVectorElements is sympifiable"""
        import sympy

        p_vec = ParameterVector("p", length=2)
        theta = Parameter("theta")

        expression = theta + 1
        expression = expression.subs({theta: p_vec[0]})
        result = expression.sympify()
        expected = sympy.Symbol("p[0]") + 1
        self.assertEqual(expected, result)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the 'clifford' unitary synthesis plugin."""

import math

from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library import UnitaryGate
from qiskit.converters import dag_to_circuit
from qiskit.quantum_info import Operator, get_clifford_gate_names
from qiskit.transpiler.passes import UnitarySynthesis, CliffordUnitarySynthesis
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestCliffordSynthesis(QiskitTestCase):
    """Test the 'clifford' unitary synthesis plugin."""

    def setUp(self):
        super().setUp()
        self._create_unitaries()

    def _create_unitaries(self):
        """
        Creates unitary gates of varying sizes, some of which can be represented
        as Cliffords and some of which cannot.
        """

        # clifford unitary over 1 qubit
        c1 = QuantumCircuit(1)
        c1.h(0)
        c1.s(0)
        self.uc1 = UnitaryGate(Operator(c1).data)

        # clifford unitary over 2 qubits
        c2 = QuantumCircuit(2)
        c2.h(0)
        c2.rz(math.pi / 4, 1)
        c2.rz(math.pi / 4, 1)
        c2.sdg(1)
        self.uc2 = UnitaryGate(Operator(c2).data)

        # clifford unitary over 3 qubits
        c3 = QuantumCircuit(3)
        c3.h(0)
        c3.cx(0, 1)
        c3.cx(1, 2)
        c3.rx(math.pi / 2, 2)
        self.uc3 = UnitaryGate(Operator(c3).data)

        # non-clifford unitary over 1 qubit
        n1 = QuantumCircuit(1)
        n1.h(0)
        n1.rz(0.8, 0)
        self.un1 = UnitaryGate(Operator(n1).data)

        # non-clifford unitary over 2 qubits
        n2 = QuantumCircuit(2)
        n2.h(0)
        n2.rz(math.pi / 4, 1)
        n2.sdg(1)
        self.un2 = UnitaryGate(Operator(n2).data)

        # non-clifford unitary over 3 qubits
        n3 = QuantumCircuit(3)
        n3.h(0)
        n3.cx(0, 1)
        n3.cx(1, 2)
        n3.rx(0.8, 2)
        self.un3 = UnitaryGate(Operator(n3).data)

    def test_plugin(self):
        """Test running CliffordUnitarySynthesis plugin directly."""

        plugin = CliffordUnitarySynthesis()
        clifford_gate_names = set(get_clifford_gate_names())

        for gate in [self.uc1, self.uc2, self.uc3]:
            mat = gate.to_matrix()
            out = plugin.run(mat)
            self.assertLessEqual(set(out.count_ops()), clifford_gate_names)
            self.assertEqual(Operator(dag_to_circuit(out)), Operator(mat))

        for gate in [self.un1, self.un2, self.un3]:
            out = plugin.run(gate.to_matrix())
            self.assertIsNone(out)

    def test_plugin_with_parameters(self):
        """Test that we can pass parameters to the plugin."""

        plugin = CliffordUnitarySynthesis()
        for gate, config in [(self.uc1, {"min_qubits": 2}), (self.uc3, {"max_qubits": 2})]:
            mat = gate.to_matrix()
            out = plugin.run(mat, config=config)
            self.assertIsNone(out)

    def test_unitary_synthesis(self):
        """Test running the plugin from the unitary synthesis transpiler pass."""

        clifford_gate_names = set(get_clifford_gate_names())
        qubits = [2, 0, 1]

        for gate in [self.uc1, self.uc2, self.uc3]:
            qc = QuantumCircuit(3)
            qc.append(gate, qubits[: gate.num_qubits])
            transpiled = UnitarySynthesis(method="clifford")(qc)
            transpiled_ops = transpiled.count_ops()
            self.assertEqual(transpiled_ops.get("unitary", 0), 0)
            self.assertLessEqual(set(transpiled_ops), clifford_gate_names)
            self.assertEqual(Operator(transpiled), Operator(qc))

        for gate in [self.un1, self.un2, self.un3]:
            qc = QuantumCircuit(3)
            qc.append(gate, qubits[: gate.num_qubits])
            # the circuit should be unchanged
            transpiled = UnitarySynthesis(method="clifford")(qc)
            self.assertEqual(qc, transpiled)

    def test_unitary_synthesis_with_parameters(self):
        """Test that we can pass parameters to the plugin via unitary synthesis."""
        qubits = [2, 0, 1]

        for gate, config in [(self.uc1, {"min_qubits": 2}), (self.uc3, {"max_qubits": 2})]:
            qc = QuantumCircuit(3)
            qc.append(gate, qubits[: gate.num_qubits])
            transpiled = UnitarySynthesis(method="clifford", plugin_config=config)(qc)
            # the circuit should be unchanged
            self.assertEqual(qc, transpiled)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-module-docstring

from qiskit.circuit.quantumcircuit import QuantumCircuit


def template_nct_4a_1():
    """
    Template 4a_1:

    .. code-block:: text

        q_0: ───────■─────────■──
                    │         │
        q_1: ──■────┼────■────┼──
               │    │    │    │
        q_2: ──■────■────■────■──
               │  ┌─┴─┐  │  ┌─┴─┐
        q_3: ──┼──┤ X ├──┼──┤ X ├
             ┌─┴─┐└───┘┌─┴─┐└───┘
        q_4: ┤ X ├─────┤ X ├─────
             └───┘     └───┘

    Returns:
        QuantumCircuit: template as a quantum circuit.
    """
    qc = QuantumCircuit(5)
    qc.ccx(1, 2, 4)
    qc.ccx(0, 2, 3)
    qc.ccx(1, 2, 4)
    qc.ccx(0, 2, 3)
    return qc



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test cases for circuit qpy loading and saving."""

import io
import struct

from ddt import ddt, data

from qiskit.circuit import QuantumCircuit, QuantumRegister, Qubit, Parameter, Gate, annotation
from qiskit.providers.fake_provider import GenericBackendV2
from qiskit.exceptions import QiskitError
from qiskit.qpy import dump, load, formats, get_qpy_version, QPY_COMPATIBILITY_VERSION
from qiskit.qpy.common import QPY_VERSION
from qiskit.transpiler import TranspileLayout, CouplingMap
from qiskit.compiler import transpile
from qiskit.qpy.formats import FILE_HEADER_V10_PACK, FILE_HEADER_V10, FILE_HEADER_V10_SIZE
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class QpyCircuitTestCase(QiskitTestCase):
    """QPY circuit testing platform."""

    def assert_roundtrip_equal(
        self, circuit, version=None, use_symengine=None, annotation_factories=None
    ):
        """QPY roundtrip equal test."""
        qpy_file = io.BytesIO()
        if use_symengine is None:
            dump(circuit, qpy_file, version=version, annotation_factories=annotation_factories)
        else:
            dump(
                circuit,
                qpy_file,
                version=version,
                use_symengine=use_symengine,
                annotation_factories=annotation_factories,
            )
        qpy_file.seek(0)
        new_circuit = load(qpy_file, annotation_factories=annotation_factories)[0]

        self.assertEqual(circuit, new_circuit)
        self.assertEqual(circuit.layout, new_circuit.layout)
        if version is not None:
            qpy_file.seek(0)
            file_version = struct.unpack("!6sB", qpy_file.read(7))[1]
            self.assertEqual(
                version,
                file_version,
                f"Generated QPY file version {file_version} does not match request version {version}",
            )


class TestVersions(QpyCircuitTestCase):
    """Test version handling in qpy."""

    def test_invalid_qpy_version(self):
        """Test a descriptive exception is raised if QPY version is too new."""
        with io.BytesIO() as buf:
            buf.write(
                struct.pack(formats.FILE_HEADER_PACK, b"QISKIT", QPY_VERSION + 4, 42, 42, 1, 2)
            )
            buf.seek(0)
            with self.assertRaisesRegex(QiskitError, str(QPY_VERSION + 4)):
                load(buf)

    def test_get_qpy_version(self):
        """Test the get_qpy_version function."""

        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)

        for version in range(QPY_COMPATIBILITY_VERSION, QPY_VERSION + 1):
            with io.BytesIO() as qpy_file:
                dump(qc, qpy_file, version=version)
                qpy_file.seek(0)
                file_version = get_qpy_version(qpy_file)
            self.assertEqual(version, file_version)

    def test_get_qpy_version_read(self):
        """Ensure we don't advance the cursor exiting the get_qpy_version() function."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        with io.BytesIO() as qpy_file:
            dump(qc, qpy_file)
            qpy_file.seek(0)
            file_version = get_qpy_version(qpy_file)
            self.assertEqual(file_version, QPY_VERSION)
            res = load(qpy_file)[0]
        self.assertEqual(res, qc)


@ddt
class TestLayout(QpyCircuitTestCase):
    """Test circuit serialization for layout preservation."""

    @data(0, 1, 2, 3)
    def test_transpile_layout(self, opt_level):
        """Test layout preserved after transpile."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        qc.measure_all()
        backend = GenericBackendV2(num_qubits=127, coupling_map=CouplingMap.from_line(127), seed=42)
        tqc = transpile(qc, backend, optimization_level=opt_level)
        self.assert_roundtrip_equal(tqc)

    @data(0, 1, 2, 3)
    def test_transpile_with_routing(self, opt_level):
        """Test full layout with routing is preserved."""
        qc = QuantumCircuit(5)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.measure_all()
        backend = GenericBackendV2(num_qubits=127, coupling_map=CouplingMap.from_line(127), seed=42)
        tqc = transpile(qc, backend, optimization_level=opt_level)
        self.assert_roundtrip_equal(tqc)

    @data(0, 1, 2, 3)
    def test_transpile_layout_explicit_None_final_layout(self, opt_level):
        """Test layout preserved after transpile."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        qc.measure_all()
        backend = GenericBackendV2(num_qubits=127, coupling_map=CouplingMap.from_line(127), seed=42)
        tqc = transpile(qc, backend, optimization_level=opt_level)
        tqc.layout.final_layout = None
        self.assert_roundtrip_equal(tqc)

    def test_empty_layout(self):
        """Test an empty layout is preserved correctly."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        qc.measure_all()
        qc._layout = TranspileLayout(None, None, None)
        self.assert_roundtrip_equal(qc)

    def test_overlapping_definitions(self):
        """Test serialization of custom gates with overlapping definitions."""

        class MyParamGate(Gate):
            """Custom gate class with a parameter."""

            def __init__(self, phi):
                super().__init__("my_gate", 1, [phi])

            def _define(self):
                qc = QuantumCircuit(1)
                qc.rx(self.params[0], 0)
                self.definition = qc

        theta = Parameter("theta")
        two_theta = 2 * theta

        qc = QuantumCircuit(1)
        qc.append(MyParamGate(1.1), [0])
        qc.append(MyParamGate(1.2), [0])
        qc.append(MyParamGate(3.14159), [0])
        qc.append(MyParamGate(theta), [0])
        qc.append(MyParamGate(two_theta), [0])
        with io.BytesIO() as qpy_file:
            dump(qc, qpy_file)
            qpy_file.seek(0)
            new_circ = load(qpy_file)[0]
        # Custom gate classes are lowered to Gate to avoid arbitrary code
        # execution on deserialization. To compare circuit equality we
        # need to go instruction by instruction and check that they're
        # equivalent instead of doing a circuit equality check
        for new_inst, old_inst in zip(new_circ.data, qc.data):
            new_gate = new_inst.operation
            old_gate = old_inst.operation
            self.assertIsInstance(new_gate, Gate)
            self.assertEqual(new_gate.name, old_gate.name)
            self.assertEqual(new_gate.params, old_gate.params)
            self.assertEqual(new_gate.definition, old_gate.definition)

    @data(0, 1, 2, 3)
    def test_custom_register_name(self, opt_level):
        """Test layout preserved with custom register names."""
        qr = QuantumRegister(5, name="abc123")
        qc = QuantumCircuit(qr)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.measure_all()
        backend = GenericBackendV2(num_qubits=127, coupling_map=CouplingMap.from_line(127), seed=42)
        tqc = transpile(qc, backend, optimization_level=opt_level)
        self.assert_roundtrip_equal(tqc)

    @data(0, 1, 2, 3)
    def test_no_register(self, opt_level):
        """Test layout preserved with no register."""
        qubits = [Qubit(), Qubit()]
        qc = QuantumCircuit(qubits)
        qc.h(0)
        qc.cx(0, 1)
        qc.measure_all()
        backend = GenericBackendV2(num_qubits=127, coupling_map=CouplingMap.from_line(127), seed=42)
        tqc = transpile(qc, backend, optimization_level=opt_level)
        # Manually validate to deal with qubit equality needing exact objects
        qpy_file = io.BytesIO()
        dump(tqc, qpy_file)
        qpy_file.seek(0)
        new_circuit = load(qpy_file)[0]
        self.assertEqual(tqc, new_circuit)
        initial_layout_old = tqc.layout.initial_layout.get_physical_bits()
        initial_layout_new = new_circuit.layout.initial_layout.get_physical_bits()
        for i in initial_layout_old:
            self.assertIsInstance(initial_layout_old[i], Qubit)
            self.assertIsInstance(initial_layout_new[i], Qubit)
            if initial_layout_old[i]._register is not None:
                self.assertEqual(initial_layout_new[i], initial_layout_old[i])
            else:
                self.assertIsNone(initial_layout_new[i]._register)
                self.assertIsNone(initial_layout_old[i]._index)
                self.assertIsNone(initial_layout_new[i]._index)
        self.assertEqual(
            list(tqc.layout.input_qubit_mapping.values()),
            list(new_circuit.layout.input_qubit_mapping.values()),
        )
        self.assertEqual(tqc.layout.final_layout, new_circuit.layout.final_layout)


class TestVersionArg(QpyCircuitTestCase):
    """Test explicitly setting a qpy version in dump()."""

    def test_invalid_version_value(self):
        """Assert we raise an error with an invalid version request."""
        qc = QuantumCircuit(2)
        with self.assertRaises(ValueError):
            dump(qc, io.BytesIO(), version=3)

    def test_compatibility_version_roundtrip(self):
        """Test the version is set correctly when specified."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        qc.measure_all()
        self.assert_roundtrip_equal(qc, version=QPY_COMPATIBILITY_VERSION)

    def test_nested_params_subs(self):
        """Test substitution works."""
        qc = QuantumCircuit(1)
        a = Parameter("a")
        b = Parameter("b")
        expr = a + b
        expr = expr.subs({b: a})
        qc.ry(expr, 0)
        self.assert_roundtrip_equal(qc)

    def test_all_the_expression_ops(self):
        """Test a circuit with an expression that uses all the ops available."""
        qc = QuantumCircuit(1)
        a = Parameter("a")
        b = Parameter("b")
        c = Parameter("c")
        d = Parameter("d")

        expression = (a + b.sin() / 4) * c**2
        final_expr = (
            (expression.cos() + d.arccos() - d.arcsin() + d.arctan() + d.tan()) / d.exp()
            + expression.gradient(a)
            + expression.log()
            - a.sin()
            - b.conjugate()
        )
        final_expr = final_expr.abs()
        final_expr = final_expr.subs({c: a})

        qc.rx(final_expr, 0)
        self.assert_roundtrip_equal(qc)

    def test_rpow(self):
        """Test rpow works as expected"""
        qc = QuantumCircuit(1)
        a = Parameter("A")
        b = Parameter("B")
        expr = 3.14159**a
        expr = expr**b
        expr = 1.2345**expr
        qc.ry(expr, 0)
        self.assert_roundtrip_equal(qc)

    def test_rsub(self):
        """Test rsub works as expected"""
        qc = QuantumCircuit(1)
        a = Parameter("A")
        b = Parameter("B")
        expr = 3.14159 - a
        expr = expr - b
        expr = 1.2345 - expr
        qc.ry(expr, 0)
        self.assert_roundtrip_equal(qc)

    def test_rdiv(self):
        """Test rdiv works as expected"""
        qc = QuantumCircuit(1)
        a = Parameter("A")
        b = Parameter("B")
        expr = 3.14159 / a
        expr = expr / b
        expr = 1.2345 / expr
        qc.ry(expr, 0)
        self.assert_roundtrip_equal(qc)


@ddt
class TestUseSymengineFlag(QpyCircuitTestCase):
    """Test that the symengine flag works correctly."""

    @data(True, False)
    def test_use_symengine_with_bool_like(self, use_symengine):
        """Test that the use_symengine flag is set correctly with a bool-like input."""

        class Booly:  # pylint: disable=missing-class-docstring,missing-function-docstring
            def __init__(self, value):
                self.value = value

            def __bool__(self):
                return self.value

        theta = Parameter("theta")
        two_theta = 2 * theta
        qc = QuantumCircuit(1)
        qc.rx(two_theta, 0)
        qc.measure_all()
        # Assert Roundtrip works
        # `use_symengine` is near-completely ignored with QPY versions 13+; it doesn't actually
        # matter if we _have_ symengine installed or not, because those QPYs don't ever use it
        # (except for setting a single byte in the header, which is promptly ignored).
        self.assert_roundtrip_equal(qc, use_symengine=Booly(use_symengine), version=13)
        # Also check the qpy symbolic expression encoding is correct in the
        # payload
        with io.BytesIO() as file_obj:
            dump(qc, file_obj, use_symengine=Booly(use_symengine))
            file_obj.seek(0)
            header_data = FILE_HEADER_V10._make(
                struct.unpack(
                    FILE_HEADER_V10_PACK,
                    file_obj.read(FILE_HEADER_V10_SIZE),
                )
            )
            self.assertEqual(header_data.symbolic_encoding, b"e" if use_symengine else b"p")


class TestSymbolExpr(QpyCircuitTestCase):
    """Test QPY with SymbolExpr"""

    def test_back_slash(self):
        """Test Parameter with back slash"""
        qc = QuantumCircuit(2)
        alpha = Parameter(r"\alpha")
        beta = Parameter(r"\beta")
        qc.rz(alpha + beta, 0)
        self.assert_roundtrip_equal(qc)

    def test_math_mode_backslash(self):
        """Test Parameter with mathmode backslah."""
        qc = QuantumCircuit(2)
        alpha = Parameter(r"$\alpha$")
        beta = Parameter(r"$\beta$")
        qc.rz(alpha + beta, 0)
        self.assert_roundtrip_equal(qc)


class TestAnnotations(QpyCircuitTestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring,redefined-outer-name

    def test_wrapping_openqasm3(self):
        class My(annotation.Annotation):
            def __init__(self, namespace, value):
                self.namespace = namespace
                self.value = value

            def __eq__(self, other):
                return (
                    isinstance(other, My)
                    and self.namespace == other.namespace
                    and self.value == other.value
                )

        class Serializer(annotation.OpenQASM3Serializer):
            def load(self, namespace, payload):
                return My(namespace, payload)

            def dump(self, annotation):
                return annotation.value

        qc = QuantumCircuit()
        with qc.box([My("my.a", "hello"), My("my.b", "world")]):
            with qc.box([My("my.c", "")]):
                pass
        self.assert_roundtrip_equal(qc, annotation_factories={"my": Serializer().as_qpy()})

    def test_simple_serializer(self):
        outer_self = self

        class Dummy(annotation.Annotation):
            namespace = "dummy"

            def __eq__(self, other):
                return isinstance(other, Dummy)

        class Serializer(annotation.QPYSerializer):
            def load_annotation(self, payload):
                outer_self.assertEqual(payload, b"SENTINEL")
                return Dummy()

            def dump_annotation(self, namespace, annotation):
                outer_self.assertEqual(namespace, "dummy")
                return b"SENTINEL"

        qc = QuantumCircuit()
        with qc.box([Dummy()]):
            pass
        self.assert_roundtrip_equal(qc, annotation_factories={"dummy": Serializer})

    def test_stateful_serializer(self):
        outer_self = self

        class My(annotation.Annotation):
            namespace = "my"

            def __init__(self, value):
                self.value = value

            def __eq__(self, other):
                return isinstance(other, My) and self.value == other.value

        class Serializer(annotation.QPYSerializer):
            def __init__(self):
                self.loaded_state = False
                self.dumped_state = False
                self.num_serialized = 0

            def dump_annotation(self, namespace, annotation):
                outer_self.assertFalse(self.dumped_state)
                outer_self.assertFalse(self.loaded_state)
                self.num_serialized += 1
                return annotation.value.encode("utf-8")

            def dump_state(self):
                # This is a check that there are two separate instances when moving into an inner
                # circuit; both the outer and inner `box`es have two annotations.
                outer_self.assertEqual(self.num_serialized, 2)
                outer_self.assertFalse(self.dumped_state)
                outer_self.assertFalse(self.loaded_state)
                self.dumped_state = True
                return b"SENTINEL"

            def load_annotation(self, payload):
                outer_self.assertFalse(self.dumped_state)
                outer_self.assertTrue(self.loaded_state)
                return My(payload.decode("utf-8"))

            def load_state(self, namespace, payload):
                outer_self.assertEqual(namespace, "my")
                outer_self.assertEqual(payload, b"SENTINEL")
                outer_self.assertFalse(self.dumped_state)
                outer_self.assertFalse(self.loaded_state)
                self.loaded_state = True

        qc = QuantumCircuit()
        with qc.box([My("hello"), My("world")]):
            with qc.box([My("inner"), My("number 2")]):
                pass
        self.assert_roundtrip_equal(qc, annotation_factories={"my": Serializer})

    def test_multiple_serializers(self):
        outer_self = self

        class TypeA(annotation.Annotation):
            namespace = "a"

            def __eq__(self, other):
                return isinstance(other, TypeA)

        class TypeB(annotation.Annotation):
            namespace = "b"

            def __eq__(self, other):
                return isinstance(other, TypeB)

        class SerializerA(annotation.QPYSerializer):
            def dump_annotation(self, namespace, annotation):
                outer_self.assertEqual(namespace, "a")
                outer_self.assertIsInstance(annotation, TypeA)
                return b"A"

            def dump_state(self):
                return b"STATE A"

            def load_annotation(self, payload):
                outer_self.assertEqual(payload, b"A")
                return TypeA()

            def load_state(self, namespace, payload):
                outer_self.assertEqual(namespace, "a")
                outer_self.assertEqual(payload, b"STATE A")

        class SerializerB(annotation.QPYSerializer):
            def dump_annotation(self, namespace, annotation):
                outer_self.assertEqual(namespace, "b")
                outer_self.assertIsInstance(annotation, TypeB)
                return b"B"

            def dump_state(self):
                return b"STATE B"

            def load_annotation(self, payload):
                outer_self.assertEqual(payload, b"B")
                return TypeB()

            def load_state(self, namespace, payload):
                outer_self.assertEqual(namespace, "b")
                outer_self.assertEqual(payload, b"STATE B")

        qc = QuantumCircuit()
        with qc.box([TypeA(), TypeB()]):
            with qc.box([TypeB(), TypeA()]):
                pass
        with qc.box([TypeB(), TypeA()]):
            with qc.box([TypeA()]):
                pass
        self.assert_roundtrip_equal(qc, annotation_factories={"a": SerializerA, "b": SerializerB})

    def test_parent_namespacing(self):
        outer_self = self

        class My(annotation.Annotation):
            def __init__(self, namespace, value):
                self.namespace = namespace
                self.value = value

            def __eq__(self, other):
                return (
                    isinstance(other, My)
                    and self.namespace == other.namespace
                    and self.value == other.value
                )

        triggered_not_implemented = False

        class Serializer(annotation.QPYSerializer):
            # The idea with this family of serialisers is that we say the "value" we expect to see
            # in the annotations it will handle, then don't actually store _anything_ in the QPY for
            # that.  We only store the _actual_ namespace of the annotation (not the parent we were
            # called with) in the state field.

            def __init__(self, value):
                self.actual_namespace = None
                self.value = value

            def dump_annotation(self, namespace, annotation):
                if annotation.value != self.value:
                    nonlocal triggered_not_implemented
                    triggered_not_implemented = True
                    return NotImplemented
                if self.actual_namespace is None:
                    self.actual_namespace = annotation.namespace
                else:
                    outer_self.assertEqual(annotation.namespace, self.actual_namespace)
                return b""

            def load_annotation(self, payload):
                return My(self.actual_namespace, self.value)

            def dump_state(self):
                outer_self.assertIsNotNone(self.actual_namespace)
                return self.actual_namespace.encode("utf-8") + b"\x00" + self.value.encode("utf-8")

            def load_state(self, namespace, payload):
                namespace, value = payload.split(b"\x00")
                self.actual_namespace = namespace.decode("utf-8")
                self.value = value.decode("utf-8")

        qc = QuantumCircuit()
        with qc.box(
            [My("a.b", "looks at a"), My("a.c", "looks at a.c"), My("a.d", "looks at global")]
        ):
            pass
        self.assert_roundtrip_equal(
            qc,
            annotation_factories={
                "a.c": (lambda: Serializer("looks at a.c")),
                "a": (lambda: Serializer("looks at a")),
                "": (lambda: Serializer("looks at global")),
            },
        )
        self.assertTrue(triggered_not_implemented)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test Clifford+T optimization pass"""

import numpy as np

from ddt import ddt, data

from qiskit.circuit import QuantumCircuit
from qiskit.transpiler.passes import SolovayKitaev, OptimizeCliffordT
from qiskit.quantum_info import Operator
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestOptimizeCliffordT(QiskitTestCase):
    """Test the OptimizeCliffordT pass."""

    angles = np.linspace(0, 2 * np.pi, 10)

    @data(*angles)
    def test_solovay_kitaev_rx(self, angle):
        """Test optimization of circuits coming out of the Solovay-Kitaev pass."""
        qc = QuantumCircuit(1)
        qc.rx(angle, 0)

        # Run Solovay-Kitaev pass on qc
        transpiled = SolovayKitaev()(qc)
        self.assertLessEqual(set(transpiled.count_ops()), {"h", "t", "tdg"})

        # Run Clifford+T optimization pass on the transpiled circuit
        optimized = OptimizeCliffordT()(transpiled)

        self.assertTrue(Operator(transpiled), Operator(optimized))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Instruction sub-classes for dynamic circuits."""


from ._builder_utils import condition_resources, node_resources, LegacyResources
from .control_flow import ControlFlowOp
from .continue_loop import ContinueLoopOp
from .break_loop import BreakLoopOp

from .box import BoxOp
from .if_else import IfElseOp
from .while_loop import WhileLoopOp
from .for_loop import ForLoopOp
from .switch_case import SwitchCaseOp, CASE_DEFAULT


CONTROL_FLOW_OP_NAMES = frozenset(("for_loop", "while_loop", "if_else", "switch_case", "box"))
"""Set of the instruction names of Qiskit's known control-flow operations."""


def get_control_flow_name_mapping():
    """Return a dictionary mapping the names of control-flow operations
    to their corresponding classes."

    Examples:

        .. code-block:: python

            from qiskit.circuit import get_control_flow_name_mapping

            ctrl_flow_name_map = get_control_flow_name_mapping()
            if_else_object = ctrl_flow_name_map["if_else"]

            print(if_else_object)

        .. code-block:: text

            <class 'qiskit.circuit.controlflow.if_else.IfElseOp'>
    """

    name_mapping = {
        "if_else": IfElseOp,
        "while_loop": WhileLoopOp,
        "for_loop": ForLoopOp,
        "switch_case": SwitchCaseOp,
        "box": BoxOp,
    }
    return name_mapping



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test Clifford+T transpilation pipeline"""

import numpy as np
from ddt import ddt, data

from qiskit.converters import circuit_to_dag
from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library import QFTGate, iqp, GraphStateGate
from qiskit.transpiler.passes.utils import CheckGateDirection
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from qiskit.transpiler import CouplingMap
from qiskit.providers.fake_provider import GenericBackendV2
from qiskit.quantum_info import get_clifford_gate_names

from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestCliffordTPassManager(QiskitTestCase):
    """Test Clifford+T transpilation pipeline."""

    @data(0, 1, 2, 3)
    def test_cliffords_1q(self, optimization_level):
        """Clifford+T transpilation of a circuit with single-qubit Clifford gates."""
        qc = QuantumCircuit(3)
        qc.h(0)
        qc.s(1)
        qc.h(2)
        qc.sdg(2)
        qc.h(2)
        qc.h(2)

        basis_gates = ["cx", "s", "sdg", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))
        # The resulting circuit should not have any T/Tdg-gates.
        self.assertEqual(_get_t_count(transpiled), 0)

    @data(0, 1, 2, 3)
    def test_complex_clifford(self, optimization_level):
        """Clifford+T transpilation of a more complex Clifford circuit."""
        qc = QuantumCircuit(3)
        qc.h(0)
        qc.s(0)
        qc.cx(1, 0)
        qc.cx(0, 1)
        qc.s(1)
        qc.sdg(0)
        qc.cx(1, 0)

        basis_gates = ["cx", "s", "sdg", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))
        # The resulting circuit should not have any T/Tdg-gates.
        self.assertEqual(_get_t_count(transpiled), 0)

    @data(0, 1, 2, 3)
    def test_rx(self, optimization_level):
        """Clifford+T transpilation of a circuit with a single-qubit rotation gate,
        requiring the usage of the Solovay-Kitaev decomposition.
        """
        qc = QuantumCircuit(1)
        qc.rx(0.8, 0)

        basis_gates = ["cx", "s", "sdg", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))

    @data(0, 1, 2, 3)
    def test_rx_pi2(self, optimization_level):
        """Clifford+T transpilation of a circuit with a single-qubit rotation gate
        that corresponds to a Clifford gate.
        """
        qc = QuantumCircuit(1)
        qc.rx(np.pi / 2, 0)

        basis_gates = get_clifford_gate_names() + ["t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))
        # The resulting circuit should not have any T/Tdg-gates.
        self.assertEqual(_get_t_count(transpiled), 0)

    @data(0, 1, 2, 3)
    def test_rx_pi4(self, optimization_level):
        """Clifford+T transpilation of a circuit with a single-qubit rotation gate
        with a "nice" angle.
        """
        qc = QuantumCircuit(1)
        qc.rx(np.pi / 4, 0)

        basis_gates = ["cx", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))
        self.assertEqual(transpiled.count_ops(), {"h": 2, "t": 1})

    @data(0, 1, 2, 3)
    def test_qft(self, optimization_level):
        """Clifford+T transpilation of a more complex circuit, requiring the usage of the
        Solovay-Kitaev decomposition.
        """
        qc = QuantumCircuit(4)
        qc.append(QFTGate(4), [0, 1, 2, 3])

        basis_gates = ["cx", "s", "sdg", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))

    @data(0, 1, 2, 3)
    def test_iqp(self, optimization_level):
        """Clifford+T transpilation of IQP circuits."""
        interactions = np.array([[6, 5, 1], [5, 4, 3], [1, 3, 2]])
        qc = iqp(interactions)

        basis_gates = ["cx", "s", "sdg", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))

        # The transpiled circuit should be fairly efficient in terms of T/Tdg gates:
        # the circuit contains 3 powers of CZ-gates, each leading to at most 9 T/Tdg gates,
        # and 3 powers of T-gates, each leading to at most 4 T/Tdg gates.
        # Importantly, the transpilation should not make this worse.
        max_t_size = 3 * 9 + 3 * 4
        self.assertLessEqual(_get_t_count(transpiled), max_t_size)

    @data(0, 1, 2, 3)
    def test_graph_state(self, optimization_level):
        """Clifford+T transpilation of graph-state circuits."""
        adjacency_matrix = [
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
        graph_state_gate = GraphStateGate(adjacency_matrix)

        qc = QuantumCircuit(5)
        qc.append(graph_state_gate, [0, 1, 2, 3, 4])

        basis_gates = ["cx", "s", "sdg", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)
        transpiled_ops = transpiled.count_ops()
        self.assertLessEqual(set(transpiled_ops), set(basis_gates))
        # The resulting circuit should not have any T/Tdg-gates.
        self.assertEqual(_get_t_count(transpiled), 0)

    @data(0, 1, 2, 3)
    def test_ccx(self, optimization_level):
        """Clifford+T transpilation of the Toffoli circuit."""
        qc = QuantumCircuit(3)
        qc.ccx(0, 1, 2)

        basis_gates = ["cx", "h", "t", "tdg"]
        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, optimization_level=optimization_level
        )
        transpiled = pm.run(qc)

        # Should get the efficient decomposition of the Toffoli gates into Clifford+T.
        self.assertEqual(transpiled.count_ops(), {"cx": 6, "t": 4, "tdg": 3, "h": 2})

    def test_t_gates(self):
        """Clifford+T transpilation of a circuit with T/Tdg-gates."""
        qc = QuantumCircuit(3)
        qc.t(0)
        qc.tdg(1)
        qc.t(2)
        qc.tdg(2)

        basis_gates = ["h", "t", "tdg"]
        pm = generate_preset_pass_manager(basis_gates=basis_gates)
        transpiled = pm.run(qc)

        # The single T/Tdg gates on qubits 0 and 1 should remain, the T/Tdg pair on qubit 2
        # should cancel out.
        expected = QuantumCircuit(3)
        expected.t(0)
        expected.tdg(1)

        self.assertEqual(transpiled, expected)

    def test_gate_direction_remapped(self):
        """Test that gate directions are correct."""
        qc = QuantumCircuit(2)
        qc.cx(0, 1)

        basis_gates = ["cx", "h", "t", "tdg"]
        coupling_map = CouplingMap([[1, 0]])

        pm = generate_preset_pass_manager(
            basis_gates=basis_gates, coupling_map=coupling_map, optimization_level=0
        )

        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))

        # Make sure gate direction is correct
        pass_ = CheckGateDirection(coupling_map=coupling_map)
        pass_.run(circuit_to_dag(transpiled))
        self.assertTrue(pass_.property_set["is_direction_mapped"])

    @data(["t", "h"], ["tdg", "h"], ["t", "sx"], ["t", "sxdg"], ["tdg", "sx"], ["tdg", "sxdg"])
    def test_clifford_t_bases(self, basis_gates):
        """Test transpiling into various Clifford+T basis sets."""
        qc = QuantumCircuit(2)
        qc.rz(0.8, 0)
        pm = generate_preset_pass_manager(basis_gates=basis_gates, optimization_level=0)
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))

    def test_target(self):
        """Clifford+T transpilation given Target."""
        qc = QuantumCircuit(2)
        qc.rz(0.8, 0)

        basis_gates = ["cx", "t", "tdg", "h"]
        backend = GenericBackendV2(5, basis_gates=basis_gates)
        pm = generate_preset_pass_manager(target=backend.target, optimization_level=0)
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))

    def test_get_clifford_gate_names(self):
        """Test transpiling with get_clifford_gate_names."""
        qc = QuantumCircuit(1)
        qc.rx(0.8, 0)

        basis_gates = get_clifford_gate_names() + ["t", "tdg"]
        pm = generate_preset_pass_manager(basis_gates=basis_gates)
        transpiled = pm.run(qc)
        self.assertLessEqual(set(transpiled.count_ops()), set(basis_gates))


def _get_t_count(qc):
    """Returns the number of T/Tdg gates in a circuit."""
    ops = qc.count_ops()
    return ops.get("t", 0) + ops.get("tdg", 0)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
=================================
Clifford Unitary Synthesis Plugin
=================================

.. autosummary::
   :toctree: ../stubs/

   CliffordUnitarySynthesis
"""

from __future__ import annotations

import math

from qiskit.exceptions import QiskitError
from qiskit.converters import circuit_to_dag
from qiskit.quantum_info import Operator, Clifford
from qiskit.quantum_info.operators.predicates import matrix_equal

from .plugin import UnitarySynthesisPlugin


class CliffordUnitarySynthesis(UnitarySynthesisPlugin):
    """A Clifford unitary synthesis plugin.

    The plugin is invoked by the :class:`.UnitarySynthesis` transpiler pass
    when the parameter ``method`` is set to ``"clifford"``.

    The plugin checks if the given unitary can be represented by a Clifford,
    in which case it returns a circuit implementing this unitary and
    consisting only of Clifford gates.

    In addition, the parameter ``plugin_config`` of :class:`.UnitarySynthesis`
    can be used to pass the following plugin-specific parameters:

    * min_qubits: the minumum number of qubits to consider (the default value is 1).

    * max_qubits: the maximum number of qubits to consider (the default value is 3).

    """

    @property
    def max_qubits(self):
        return None

    @property
    def min_qubits(self):
        return None

    @property
    def supports_basis_gates(self):
        return False

    @property
    def supports_coupling_map(self):
        return False

    @property
    def supports_natural_direction(self):
        return False

    @property
    def supports_pulse_optimize(self):
        return False

    @property
    def supports_gate_lengths(self):
        return False

    @property
    def supports_gate_errors(self):
        return False

    @property
    def supported_bases(self):
        return None

    def run(self, unitary, **options):
        """Run the CliffordUnitarySynthesis plugin on the given unitary."""

        config = options.get("config") or {}

        min_qubits = config.get("min_qubits", 1)
        max_qubits = config.get("max_qubits", 3)

        num_qubits = int(math.log2(unitary.shape[0]))

        dag = None
        if min_qubits <= num_qubits <= max_qubits:
            try:
                # Attempts to construct a Clifford from a unitary
                # (raises a QiskitError if this is not possible)
                cliff = Clifford.from_matrix(unitary)

                circuit = cliff.to_circuit()

                # Constructing Clifford from a unitary discards the global phase,
                # so we need to reconstruct it.
                props = {}
                if not matrix_equal(
                    unitary, Operator(circuit)._data, ignore_phase=True, props=props
                ):
                    raise QiskitError("Clifford synthesis is incorrect")
                circuit.global_phase -= props["phase_difference"]

                dag = circuit_to_dag(circuit)
            except QiskitError:
                pass

        return dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=invalid-name

"""
Visualization function for DAG circuit representation.
"""

import io
import subprocess

from rustworkx.visualization import graphviz_draw

from qiskit.dagcircuit.dagnode import DAGOpNode, DAGInNode, DAGOutNode
from qiskit.dagcircuit.dagcircuit import DAGCircuit
from qiskit.circuit import Qubit, Clbit, ClassicalRegister
from qiskit.circuit.classical import expr
from qiskit.converters import dagdependency_to_circuit
from qiskit.utils import optionals as _optionals
from qiskit.exceptions import InvalidFileError
from .exceptions import VisualizationError

from .style import load_style
from .dag.dagstyle import DAGStyleDict, DAGDefaultStyle


IMAGE_TYPES = {
    "canon",
    "cmap",
    "cmapx",
    "cmapx_np",
    "dia",
    "dot",
    "fig",
    "gd",
    "gd2",
    "gif",
    "hpgl",
    "imap",
    "imap_np",
    "ismap",
    "jpe",
    "jpeg",
    "jpg",
    "mif",
    "mp",
    "pcl",
    "pdf",
    "pic",
    "plain",
    "plain-ext",
    "png",
    "ps",
    "ps2",
    "svg",
    "svgz",
    "vml",
    "vmlz",
    "vrml",
    "vtx",
    "wbmp",
    "xdor",
    "xlib",
}


@_optionals.HAS_GRAPHVIZ.require_in_call
@_optionals.HAS_PIL.require_in_call
def dag_drawer(
    dag,
    scale=0.7,
    filename=None,
    style="color",
):
    """Plot the directed acyclic graph (dag) to represent operation dependencies
    in a quantum circuit.

    This function calls the :func:`~rustworkx.visualization.graphviz_draw` function from the
    ``rustworkx`` package to draw the DAG.

    Args:
        dag (DAGCircuit or DAGDependency): The dag to draw.
        scale (float): scaling factor
        filename (str): file path to save image to (format inferred from name)
        style (dict or str): Style name, file name of style JSON file, or a
            dictionary specifying the style.

            * The supported style names are 'plain': B&W graph, 'color' (default):
                (color input/output/op nodes)
            * If given a JSON file, e.g. ``my_style.json`` or ``my_style`` (the ``.json``
                extension may be omitted), this function attempts to load the style dictionary
                from that location. Note, that the JSON file must completely specify the
                visualization specifications. The file is searched for in
                ``qiskit/visualization/circuit/styles``, the current working directory, and
                the location specified in ``~/.qiskit/settings.conf``.
            * If ``None`` the default style ``"color"`` is used or, if given, the default style
                specified in ``~/.qiskit/settings.conf``.

    Returns:
        PIL.Image: if in Jupyter notebook and not saving to file,
            otherwise None.

    Raises:
        VisualizationError: when style is not recognized.
        InvalidFileError: when filename provided is not valid
        ValueError: If the file extension for ``filename`` is not an image
            type supported by Graphviz.

    Example:
        .. plot::
            :include-source:
            :nofigs:

            from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit
            from qiskit.converters import circuit_to_dag
            from qiskit.visualization import dag_drawer

            q = QuantumRegister(3, 'q')
            c = ClassicalRegister(3, 'c')
            circ = QuantumCircuit(q, c)
            circ.h(q[0])
            circ.cx(q[0], q[1])
            circ.measure(q[0], c[0])
            with circ.if_test((c, 2)):
                circ.rz(0.5, q[1])

            dag = circuit_to_dag(circ)

            style = {
                "inputnodecolor": "pink",
                "outputnodecolor": "lightblue",
                "opnodecolor": "red",
            }

            dag_drawer(dag, style=style)
    """

    from PIL import Image

    # NOTE: use type str checking to avoid potential cyclical import
    # the two tradeoffs ere that it will not handle subclasses and it is
    # slower (which doesn't matter for a visualization function)
    type_str = str(type(dag))
    register_bit_labels = {
        bit: f"{reg.name}[{idx}]"
        for reg in list(dag.qregs.values()) + list(dag.cregs.values())
        for (idx, bit) in enumerate(reg)
    }

    graph_attrs = {}
    if isinstance(style, dict):
        for attr in ["fontsize", "bgcolor", "dpi", "pad"]:
            if attr in style:
                graph_attrs[attr] = str(style[attr])

    style, _ = load_style(
        style,
        style_dict=DAGStyleDict,
        default_style=DAGDefaultStyle(),
        user_config_opt="circuit_graphviz_style",
        user_config_path_opt="circuit_graphviz_style_path",
        raise_error_if_not_found=True,
    )

    if "DAGDependency" in type_str:
        # pylint: disable=cyclic-import
        from qiskit.visualization.circuit._utils import get_bit_reg_index

        qubit_indices = {bit: index for index, bit in enumerate(dag.qubits)}
        clbit_indices = {bit: index for index, bit in enumerate(dag.clbits)}
        graph_attrs.update({"dpi": str(100 * scale)})

        dag_dep_circ = dagdependency_to_circuit(dag)

        def node_attr_func(node):
            if "DAGDependencyV2" in type_str:
                nid_str = str(node._node_id)
            else:
                nid_str = str(node.node_id)

            n = {}
            args = []
            for count, arg in enumerate(node.qargs + node.cargs):
                if count > 4:
                    args.append("...")
                    break
                if isinstance(arg, Qubit):
                    f_str = f"q_{qubit_indices[arg]}"
                elif isinstance(arg, Clbit):
                    f_str = f"c_{clbit_indices[arg]}"
                else:
                    f_str = f"{arg.index}"
                arg_str = register_bit_labels.get(arg, f_str)
                args.append(arg_str)

            n["label"] = (
                nid_str + ": " + str(node.name) + " (" + str(args)[1:-1].replace("'", "") + ")"
            )
            if getattr(node.op, "condition", None):
                condition = node.op.condition
                if isinstance(condition, expr.Expr):
                    cond_txt = " (cond: [Expr]) ("
                elif isinstance(condition[0], ClassicalRegister):
                    cond_txt = f" (cond: {condition[0].name}, {int(condition[1])}) ("
                else:
                    register, bit_index, reg_index = get_bit_reg_index(dag_dep_circ, condition[0])
                    if register is not None:
                        cond_txt = f" (cond: {register.name}[{reg_index}], {int(condition[1])}) ("
                    else:
                        cond_txt = f" (cond: {bit_index}, {int(condition[1])}) ("
                n["label"] = (
                    nid_str
                    + ": "
                    + str(node.name)
                    + cond_txt
                    + str(args)[1:-1].replace("'", "")
                    + ")"
                )

            if isinstance(style, dict):
                n["style"] = "filled"

                if "nodecolor" in style:
                    n["fillcolor"] = style["nodecolor"]

                if "fontsize" in style:
                    n["fontsize"] = str(style["fontsize"])

                if node.name == "barrier":
                    n["fillcolor"] = style["barriercolor"]
                elif getattr(node.op, "_directive", False):
                    n["fillcolor"] = style["directivecolor"]
                elif getattr(node.op, "condition", None):
                    n["fillcolor"] = style["conditioncolor"]
                elif node.name == "measure":
                    n["fillcolor"] = style["measurecolor"]

                return n
            else:
                raise VisualizationError(f"Unrecognized style {style} for the dag_drawer.")

        edge_attr_func = None

    else:
        graph_attrs.update({"dpi": str(100 * scale)})

        def node_attr_func(node):
            n = {}
            if isinstance(node, DAGOpNode):
                n["label"] = node.name
            if isinstance(node, DAGInNode):
                if isinstance(node.wire, Qubit):
                    label = register_bit_labels.get(node.wire, f"q_{dag.find_bit(node.wire).index}")
                elif isinstance(node.wire, Clbit):
                    label = register_bit_labels.get(node.wire, f"c_{dag.find_bit(node.wire).index}")
                else:
                    label = str(node.wire.name)

                n["label"] = label
            if isinstance(node, DAGOutNode):
                if isinstance(node.wire, Qubit):
                    label = register_bit_labels.get(
                        node.wire, f"q[{dag.find_bit(node.wire).index}]"
                    )
                elif isinstance(node.wire, Clbit):
                    label = register_bit_labels.get(
                        node.wire, f"c[{dag.find_bit(node.wire).index}]"
                    )
                else:
                    label = str(node.wire.name)
                n["label"] = label

            if isinstance(style, dict):
                n["style"] = "filled"

                if "nodecolor" in style:
                    n["fillcolor"] = style["nodecolor"]

                if "fontsize" in style:
                    n["fontsize"] = str(style["fontsize"])

                if isinstance(node, DAGInNode):
                    if "inputnodecolor" in style:
                        n["fillcolor"] = style["inputnodecolor"]
                    if "inputnodefontcolor" in style:
                        n["fontcolor"] = style["inputnodefontcolor"]
                if isinstance(node, DAGOutNode):
                    if "outputnodecolor" in style:
                        n["fillcolor"] = style["outputnodecolor"]
                    if "outputnodefontcolor" in style:
                        n["fontcolor"] = style["outputnodefontcolor"]
                if isinstance(node, DAGOpNode):
                    if "opnodecolor" in style:
                        n["fillcolor"] = style["opnodecolor"]
                    if "opnodefontcolor" in style:
                        n["fontcolor"] = style["opnodefontcolor"]

                return n
            else:
                raise VisualizationError(f"Invalid style {style}")

        def edge_attr_func(edge):
            e = {}

            if isinstance(edge, Qubit):
                label = register_bit_labels.get(edge, f"q_{dag.find_bit(edge).index}")
            elif isinstance(edge, Clbit):
                label = register_bit_labels.get(edge, f"c_{dag.find_bit(edge).index}")
            else:
                label = str(edge.name)
            e["label"] = label

            if isinstance(style, dict):
                if "edgecolor" in style:
                    e["color"] = style["edgecolor"]
                if "fontsize" in style:
                    e["fontsize"] = str(style["fontsize"])

                if isinstance(edge, Qubit):
                    if "qubitedgecolor" in style:
                        e["color"] = style["qubitedgecolor"]
                if isinstance(edge, Clbit):
                    if "clbitedgecolor" in style:
                        e["color"] = style["clbitedgecolor"]
                return e

            return e

    image_type = "png"
    if filename:
        if "." not in filename:
            raise InvalidFileError("Parameter 'filename' must be in format 'name.extension'")
        image_type = filename.split(".")[-1]
        if image_type not in IMAGE_TYPES:
            raise ValueError(
                "The specified value for the image_type argument, "
                f"'{image_type}' is not a valid choice. It must be one of: "
                f"{IMAGE_TYPES}"
            )

    if isinstance(dag, DAGCircuit):
        dot_str = dag._to_dot(
            graph_attrs,
            node_attr_func,
            edge_attr_func,
        )

        prog = "dot"
        if not filename:
            dot_result = subprocess.run(
                [prog, "-T", image_type],
                input=dot_str.encode("utf-8"),
                capture_output=True,
                encoding=None,
                check=True,
                text=False,
            )
            dot_bytes_image = io.BytesIO(dot_result.stdout)
            image = Image.open(dot_bytes_image)
            return image
        else:
            subprocess.run(
                [prog, "-T", image_type, "-o", filename],
                input=dot_str,
                check=True,
                encoding="utf8",
                text=True,
            )
            return None

    else:
        return graphviz_draw(
            dag._multi_graph,
            node_attr_func,
            edge_attr_func,
            graph_attrs,
            filename,
            image_type,
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""This module contains common utils for disjoint coupling maps."""
from __future__ import annotations
from typing import Union

from qiskit.dagcircuit.dagcircuit import DAGCircuit
from qiskit.transpiler.coupling import CouplingMap
from qiskit.transpiler.target import Target
from qiskit.transpiler.exceptions import TranspilerError


def require_layout_isolated_to_component(
    dag: DAGCircuit, components_source: Union[Target, CouplingMap]
):
    """
    Check that the layout of the dag does not require connectivity across connected components
    in the CouplingMap

    Args:
        dag: DAGCircuit to check.
        components_source: Target to check against.

    Raises:
        TranspilerError: Chosen layout is not valid for the target disjoint connectivity.
    """
    if isinstance(components_source, Target):
        coupling_map = components_source.build_coupling_map(filter_idle_qubits=True)
    else:
        coupling_map = components_source
    component_sets = [set(x.graph.nodes()) for x in coupling_map.connected_components()]
    for inst in dag.two_qubit_ops():
        component_index = None
        for i, component_set in enumerate(component_sets):
            if dag.find_bit(inst.qargs[0]).index in component_set:
                component_index = i
                break
        if dag.find_bit(inst.qargs[1]).index not in component_sets[component_index]:
            raise TranspilerError(
                "The circuit has an invalid layout as two qubits need to interact in disconnected "
                "components of the coupling map. The physical qubit "
                f"{dag.find_bit(inst.qargs[1]).index} needs to interact with the "
                f"qubit {dag.find_bit(inst.qargs[0]).index} and they belong to different components"
            )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""
ND-Array container class for Estimator observables.
"""
from __future__ import annotations

from copy import deepcopy
from collections.abc import Iterable, Mapping as _Mapping
from typing import Union, Mapping, overload, TYPE_CHECKING

import numpy as np
from numpy.typing import ArrayLike

from qiskit.quantum_info import Pauli, PauliList, SparsePauliOp, SparseObservable

from .object_array import object_array
from .shape import ShapedMixin, shape_tuple


if TYPE_CHECKING:
    from qiskit.transpiler.layout import TranspileLayout


# Public API classes
__all__ = ["ObservableLike", "ObservablesArrayLike"]

IndexType = Union[int, slice, None]  # pylint: disable=used-before-assignment

ObservableLike = Union[
    str,
    Pauli,
    SparsePauliOp,
    SparseObservable,
    Mapping[Union[str, Pauli], float],
]
"""Types that can be natively used to construct a Hermitian Estimator observable."""


ObservablesArrayLike = Union[ObservableLike, ArrayLike]
"""Types that can be natively converted to an array of Hermitian Estimator observables."""


class ObservablesArray(ShapedMixin):
    """An ND-array of Hermitian observables for an :class:`.Estimator` primitive."""

    __slots__ = ("_array", "_shape")

    def __init__(
        self,
        observables: ObservablesArrayLike,
        num_qubits: int | None = None,
        copy: bool = True,
        validate: bool = True,
    ):
        """Initialize an observables array.

        Args:
            observables: An array-like of basis observable compatible objects.
            copy: Specify the ``copy`` kwarg of the :func:`.object_array` function
                when initializing observables.
            num_qubits: The number of qubits of the observables. If not specified, the number of
                qubits will be inferred from the observables. If specified, then the specified
                number of qubits must match the number of qubits in the observables.
            validate: If true, coerce entries into the internal format and validate them. If false,
                the input should already be an array-like.

        Raises:
            ValueError: If ``validate=True`` and the input observables array is not valid.
        """
        super().__init__()
        if isinstance(observables, ObservablesArray):
            observables = observables._array
        self._array = object_array(observables, copy=copy, list_types=(PauliList,))
        self._shape = self._array.shape
        self._num_qubits = num_qubits

        if validate:
            for ndi, obs in np.ndenumerate(self._array):
                basis_obs = self.coerce_observable(obs)
                if self._num_qubits is None:
                    self._num_qubits = basis_obs.num_qubits
                elif self._num_qubits != basis_obs.num_qubits:
                    raise ValueError(
                        "The number of qubits must be the same for all observables in the "
                        "observables array."
                    )
                self._array[ndi] = basis_obs
        elif self._num_qubits is None and self._array.size > 0:
            self._num_qubits = self._array.reshape(-1)[0].num_qubits

        # can happen for empty arrays
        if self._num_qubits is None:
            self._num_qubits = 0

    @staticmethod
    def _obs_to_dict(obs: SparseObservable) -> Mapping[str, float]:
        """Convert a sparse observable to a mapping from Pauli strings to coefficients."""
        result = {}
        for sparse_pauli_str, pauli_qubits, coeff in obs.to_sparse_list():

            if len(sparse_pauli_str) == 0:
                full_pauli_str = "I" * obs.num_qubits
            else:
                sorted_lists = sorted(zip(pauli_qubits, sparse_pauli_str))
                string_fragments = []
                prev_qubit = -1
                for qubit, pauli in sorted_lists:
                    string_fragments.append("I" * (qubit - prev_qubit - 1) + pauli)
                    prev_qubit = qubit

                string_fragments.append("I" * (obs.num_qubits - max(pauli_qubits) - 1))
                full_pauli_str = "".join(string_fragments)[::-1]

            # We know that the dictionary doesn't contain yet full_pauli_str as a key
            # because the observable is guaranteed to be simplified
            result[full_pauli_str] = np.real(coeff)

        return result

    def __repr__(self):
        prefix = f"{type(self).__name__}("
        suffix = f", shape={self.shape})"
        array = np.array2string(self.__array__(), prefix=prefix, suffix=suffix, threshold=50)
        return prefix + array + suffix

    def tolist(self) -> list | ObservableLike:
        """Convert to a nested list.

        Similar to Numpy's ``tolist`` method, the level of nesting
        depends on the dimension of the observables array. In the
        case of dimension 0 the method returns a single observable
        (``dict`` in the case of a weighted sum of Paulis) instead of a list.

        Examples::
            Return values for a one-element list vs one element:

                >>> from qiskit.primitives.containers.observables_array import ObservablesArray
                >>> oa = ObservablesArray.coerce(["Z"])
                >>> print(type(oa.tolist()))
                <class 'list'>
                >>> oa = ObservablesArray.coerce("Z")
                >>> print(type(oa.tolist()))
                <class 'dict'>
        """
        return self.__array__().tolist()

    def __array__(self, dtype=None, copy=None) -> np.ndarray:  # pylint: disable=unused-argument
        """Convert to a Numpy.ndarray with elements of type dict."""
        if dtype is None or dtype == object:
            tmp_result = self.__getitem__(tuple(slice(None) for _ in self._array.shape))
            if len(self._array.shape) == 0:
                result = np.ndarray(shape=self._array.shape, dtype=dict)
                result[()] = tmp_result
            else:
                result = np.ndarray(tmp_result.shape, dtype=dict)
                for ndi, obs in np.ndenumerate(tmp_result._array):
                    result[ndi] = self._obs_to_dict(obs)
            return result
        raise ValueError("Type must be 'None' or 'object'")

    def sparse_observables_array(self, copy: bool = False) -> np.ndarray:
        """Convert to a :class:`numpy.ndarray` with elements of type :class:`~.SparseObservable`.

        Args:
            copy: Whether to make a new array instance with new sparse observables as elements.

        Returns:
            A :class:`numpy.ndarray` with elements of type :class:`~.SparseObservable`.
        """
        obs = self.copy() if copy else self
        return obs._array

    @overload
    def __getitem__(self, args: int | tuple[int, ...]) -> Mapping[str, float]: ...

    @overload
    def __getitem__(self, args: IndexType | tuple[IndexType, ...]) -> ObservablesArray: ...

    def __getitem__(self, args):
        item = self._array[args]
        if not isinstance(item, np.ndarray):
            return self._obs_to_dict(item)

        return ObservablesArray(item, copy=False, validate=False)

    @overload
    def slice(self, args: int | tuple[int, ...]) -> SparseObservable: ...

    @overload
    def slice(self, args: IndexType | tuple[IndexType, ...]) -> ObservablesArray: ...

    def slice(self, args):
        """Take a slice of the observables in this array.

        .. note::
           This method does not copy observables; modifying the returned observables will affect this
           instance.

        Returns:
            A single :class:`~.SparseObservable` if an integer is given for every array axis, otherwise,
            a new :class:`~.ObservablesArray`.
        """
        item = self._array[args]
        if not isinstance(item, np.ndarray):
            return item

        return ObservablesArray(item, copy=False, validate=False)

    def reshape(self, *shape: int | Iterable[int]) -> ObservablesArray:
        """Return a new array with a different shape.

        This results in a new view of the same arrays.

        Args:
            shape: The shape of the returned array.

        Returns:
            A new array.
        """
        shape = shape_tuple(*shape)
        return ObservablesArray(self._array.reshape(shape), copy=False, validate=False)

    def ravel(self) -> ObservablesArray:
        """Return a new array with one dimension.

        The returned array has a :attr:`shape` given by ``(size, )``, where
        the size is the :attr:`~size` of this array.

        Returns:
            A new flattened array.
        """
        return self.reshape(self.size)

    @property
    def num_qubits(self) -> int:
        """The number of qubits each observable acts on."""
        return self._num_qubits

    @classmethod
    def coerce_observable(cls, observable: ObservableLike) -> SparseObservable:
        """Format an observable-like object into the internal format.

        Args:
            observable: The observable-like to format.

        Returns:
            The coerced observable.

        Raises:
            TypeError: If the input cannot be formatted because its type is not valid.
            ValueError: If the input observable is invalid.
        """
        # Pauli-type conversions
        if isinstance(observable, SparsePauliOp):
            observable = SparseObservable.from_sparse_pauli_op(observable)
        elif isinstance(observable, Pauli):
            observable = SparseObservable.from_pauli(observable)
        elif isinstance(observable, str):
            observable = SparseObservable.from_label(observable)
        elif isinstance(observable, _Mapping):
            term_list = []
            for basis, coeff in observable.items():
                if isinstance(basis, str):
                    term_list.append((basis, coeff))
                elif isinstance(basis, Pauli):
                    unphased_basis, phase = basis[:].to_label(), basis.phase
                    term_list.append((unphased_basis, complex(0, 1) ** phase * coeff))
                else:
                    raise TypeError(f"Invalid observable basis type: {type(basis)}")
            observable = SparseObservable.from_list(term_list)

        if isinstance(observable, SparseObservable):
            # Check that the operator has real coeffs
            coeffs = np.real_if_close(observable.coeffs)
            if np.iscomplexobj(coeffs):
                raise ValueError(
                    "Non-Hermitian input observable: the input SparsePauliOp has non-zero"
                    " imaginary part in its coefficients."
                )

            return SparseObservable.from_raw_parts(
                observable.num_qubits,
                coeffs,
                observable.bit_terms,
                observable.indices,
                observable.boundaries,
            ).simplify(tol=0)

        raise TypeError(f"Invalid observable type: {type(observable)}")

    @classmethod
    def coerce(cls, observables: ObservablesArrayLike) -> ObservablesArray:
        """Coerce ObservablesArrayLike into ObservableArray.

        Args:
            observables: an object to be observables array.

        Returns:
            A coerced observables array.
        """
        if isinstance(observables, ObservablesArray):
            return observables
        return cls(observables)

    def equivalent(self, other: ObservablesArray, tol: float = 1e-08) -> bool:
        """Compute whether the observable arrays are equal within a given tolerance.

        Args:
            other: Another observables array to compare with.
            tol: The tolerance to provide to :attr:`~.SparseObservable.simplify` during checking.

        Returns:
            Whether the two observables arrays have the same shape and number of qubits,
            and if so, whether they are equal within tolerance.
        """
        if self.num_qubits != other.num_qubits or self.shape != other.shape:
            return False

        zero_obs = SparseObservable.zero(self.num_qubits)
        for obs1, obs2 in zip(self._array.ravel(), other._array.ravel()):
            if (obs1 - obs2).simplify(tol) != zero_obs:
                return False

        return True

    def copy(self):
        """Return a deep copy of the array."""
        return deepcopy(self)

    def apply_layout(
        self, layout: TranspileLayout | list[int] | None, num_qubits: int | None = None
    ) -> ObservablesArray:
        """Apply a transpiler layout to this :class:`~.ObservablesArray`.

        Args:
            layout: Either a :class:`~.TranspileLayout`, a list of integers or None.
                    If both layout and ``num_qubits`` are none, a deep copy of the array is
                    returned.
            num_qubits: The number of qubits to expand the array to. If not
                provided then if ``layout`` is a :class:`~.TranspileLayout` the
                number of the transpiler output circuit qubits will be used by
                default. If ``layout`` is a list of integers the permutation
                specified will be applied without any expansion. If layout is
                None, the array will be expanded to the given number of qubits.

        Returns:
            A new :class:`.ObservablesArray` with the provided layout applied.

        Raises:
            QiskitError: ...
        """
        if layout is None and num_qubits is None:
            return self.copy()

        new_arr = np.ndarray(self.shape, dtype=SparseObservable)
        for ndi, obs in np.ndenumerate(self._array):
            new_arr[ndi] = obs.apply_layout(layout, num_qubits)

        return ObservablesArray(new_arr, validate=False)

    def validate(self):
        """Validate the consistency in observables array."""
        for obs in self._array.reshape(-1):
            if obs.num_qubits != self.num_qubits:
                raise ValueError(
                    "An observable was detected, whose number of qubits"
                    " does not match the array's number of qubits"
                )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the TranslateParameterizedGates pass"""

import unittest

from qiskit.circuit import ParameterVector, Parameter, Gate, QuantumCircuit
from qiskit.circuit.library import n_local
from qiskit.exceptions import QiskitError
from qiskit.transpiler.passes import TranslateParameterizedGates
from qiskit.providers.fake_provider import GenericBackendV2
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestTranslateParameterized(QiskitTestCase):
    """Test the pass to translate parameterized gates, but keep others as is."""

    def test_only_parameterized_is_unrolled(self):
        """Test only parameterized gates are unrolled."""
        x = ParameterVector("x", 4)
        block1 = QuantumCircuit(1)
        block1.rx(x[0], 0)

        sub_block = QuantumCircuit(2)
        sub_block.cx(0, 1)
        sub_block.rz(x[2], 0)

        block2 = QuantumCircuit(2)
        block2.ry(x[1], 0)
        block2.append(sub_block.to_gate(), [0, 1])

        block3 = QuantumCircuit(3)
        block3.ccx(0, 1, 2)

        circuit = QuantumCircuit(3)
        circuit.append(block1.to_gate(), [1])
        circuit.append(block2.to_gate(), [0, 1])
        circuit.append(block3.to_gate(), [0, 1, 2])
        circuit.cry(x[3], 0, 2)

        supported_gates = ["rx", "ry", "rz", "cp", "crx", "cry", "crz"]
        unroller = TranslateParameterizedGates(supported_gates)
        unrolled = unroller(circuit)

        expected = QuantumCircuit(3)
        expected.rx(x[0], 1)
        expected.ry(x[1], 0)
        expected.cx(0, 1)
        expected.rz(x[2], 0)
        expected.append(block3.to_gate(), [0, 1, 2])
        expected.cry(x[3], 0, 2)

        self.assertEqual(unrolled, expected)

    def test_target(self):
        """Test unrolling with a target."""
        target = GenericBackendV2(num_qubits=5).target
        circuit = n_local(2, "rz", "cx", reps=2, entanglement="linear")

        translator = TranslateParameterizedGates(target=target)
        translated = translator(circuit)

        expected_ops = {"cx": 2, "rz": 6}

        self.assertEqual(translated.count_ops(), expected_ops)

    def test_no_supported_gates_or_target(self):
        """Test an error is raised if neither of ``supported_gates`` and ``target`` is supported."""
        with self.assertRaises(ValueError):
            _ = TranslateParameterizedGates(supported_gates=None, target=None)

    def test_translation_impossible(self):
        """Test translating a parameterized gate without definition does not work."""

        x = Parameter("x")
        mission_actually_impossible = Gate("mysterious", 1, [x])

        circuit = QuantumCircuit(1)
        circuit.rx(x, 0)
        circuit.append(mission_actually_impossible, [0])

        translator = TranslateParameterizedGates(["rx", "ry", "rz"])

        with self.assertRaises(QiskitError):
            _ = translator(circuit)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for the context-aware dynamical decoupling passes."""

import math
import itertools
import contextlib
from ddt import ddt, data, unpack
import numpy as np
import rustworkx as rx

from qiskit.circuit import QuantumCircuit, Delay, Reset
from qiskit.circuit.library import SXGate, CXGate, XGate, CZGate, ECRGate
from qiskit.transpiler import PassManager, Target, InstructionProperties
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from qiskit.transpiler.passes import (
    SetLayout,
    FullAncillaAllocation,
    EnlargeWithAncilla,
    ContextAwareDynamicalDecoupling,
    ALAPScheduleAnalysis,
    ASAPScheduleAnalysis,
    ApplyLayout,
)
from qiskit.providers.fake_provider import GenericBackendV2

from test import QiskitTestCase  # pylint: disable=wrong-import-order


class Mock127q(GenericBackendV2):
    """A fake 127-qubit backend.

    The node number equals ``ibm_sherbrooke`` though the directions of the ECR gates might
    not match.

    Used instead of the GenericBackendV2 to ensure the 2-qubit gate lengths are larger
    than 2 X gates, and fixes the seed.
    """

    def __init__(self):
        coupling_map = []
        line_lengths = [13] + 5 * [14] + [13]
        lines = []
        start = 0
        for length in line_lengths:
            line = list(range(start, start + length + 1))
            coupling_map += list(zip(line[:-1], line[1:]))
            lines.append(line)
            start += length + 5  # 4 nodes are in between the lines

        for i, (prev, after) in enumerate(zip(lines[:-1], lines[1:])):
            from_front = i % 2 == 0
            for j in range(4):
                node = prev[-1] + j + 1 if from_front else prev[-1] + 4 - j  # node number
                index = 4 * j if from_front else ~(4 * j)
                coupling_map.append((prev[index], node))
                coupling_map.append((node, after[index]))

        basis_gates = ["id", "rz", "sx", "x", "ecr"]

        super().__init__(num_qubits=127, coupling_map=coupling_map, basis_gates=basis_gates, seed=9)

    def _get_noise_defaults(self, name, num_qubits):
        # gate durations and errors as (min duration, max duration, min error, max error)
        if name == "ecr":
            return (341e-9, 881e-9, 1e-3, 3e-3)
        elif name == "x":
            return (1e-10, 1.1e-10, 9e-5, 6e-3)
        elif name == "sx":
            return (5e-11, 5.5e-11, 9e-5, 6e-3)
        return super()._get_noise_defaults(name, num_qubits)


@ddt
class TestContextAwareDD(QiskitTestCase):
    """Test context-aware dynamical decoupling.

    See the reference: https://arxiv.org/abs/2403.06852v2.
    """

    def setUp(self):
        super().setUp()

        # gate times in terms of dt
        self.dt = 1e-9
        self.t_cx = 1e3
        self.t_sx = 10
        self.t_x = 20

    @contextlib.contextmanager
    def assertMultiDelayInserted(self, expected_blocks):
        """Assert that multi-delays are inserted.

        These are to be given as [(start1, end1, qubits1), (start2, end2, qubits2), ...]
        per adjacent block.
        """
        module = "qiskit.transpiler.passes.scheduling.padding.context_aware_dynamical_decoupling"
        try:
            with self.assertLogs(module, level="DEBUG") as logs:
                yield logs
        finally:
            template = "MultiDelay({}:{} on {})"  # MultiDelay(start:end on qubits)
            for delay_block in expected_blocks:
                message = f"DEBUG:{module}:Split adjacent delay block into " + "\n".join(
                    [template.format(start, end, qubits) for start, end, qubits in delay_block]
                )

                self.assertIn(message, logs.output)

    def test_full(self):
        """Test the full workflow on a simple circuit and a concrete reference."""
        circuit = QuantumCircuit(5)
        circuit.sx(circuit.qubits)
        circuit.barrier()
        circuit.cx(1, 2)
        circuit.barrier()
        circuit.cx(0, 1)
        circuit.cx(3, 4)

        target = get_toy_target(num_qubits=5, t_cx=self.t_cx)
        durations = target.durations()

        pm = _get_schedule_pm(target, list(range(circuit.num_qubits)))
        pm.append(ContextAwareDynamicalDecoupling(target))

        ref = QuantumCircuit(5)
        ref.sx(ref.qubits)
        ref.barrier()
        apply_delay_sequence(ref, 0, self.t_cx, durations, order=1)  # ctrl sequence
        ref.cx(1, 2)
        apply_delay_sequence(ref, 3, self.t_cx, durations, order=0)  # tgt sequence
        apply_delay_sequence(ref, 4, self.t_cx, durations, order=1)  # lowest seq. orthogonal to tgt

        ref.barrier()
        ref.cx(0, 1)
        apply_delay_sequence(ref, 2, self.t_cx, durations, order=2)  # ctrl+tgt sequence
        ref.cx(3, 4)

        circuit = pm.run(circuit)
        self.assertEqual(circuit, ref)

    @data(True, False)
    def test_skip_initial_delays(self, skip_initial):
        """Test initial delays are skipped.

           delays here since not after a reset
                    v
        q_0: ──■───────
             ┌─┴─┐
        q_1: ┤ X ├──■──
             └───┘┌─┴─┐
        q_2: ─────┤ X ├
               ^  └───┘
             no delay since after qubit initialization

        """
        circuit = QuantumCircuit(3)
        circuit.cx(0, 1)
        circuit.cx(1, 2)

        target = get_toy_target(num_qubits=5)
        pm = _get_schedule_pm(target, list(range(circuit.num_qubits)))
        pm.append(ContextAwareDynamicalDecoupling(target, skip_reset_qubits=skip_initial))

        circuit = pm.run(circuit)

        # a single ctrl-specific decoupling sequence with 2 X gates if initial delays are skipped,
        # otherwise we get an additional 5 sequences à 2 X gates due to the delay blocks on the
        # additional idle qubits of the 5-qubit target
        expected_x = 2 if skip_initial else 12

        self.assertEqual(circuit.count_ops().get("x", 0), expected_x)

    def test_min_duration_default(self):
        """Test the default value for minimum durations."""
        circuit = QuantumCircuit(3)
        circuit.cx(0, 1)
        circuit.cx(1, 2)

        # target = get_toy_target(num_qubits=5)
        target, max_diff = get_varying_target(5)
        dd = ContextAwareDynamicalDecoupling(target)

        self.assertAlmostEqual(2 * max_diff / target.dt, dd._min_duration)

    def test_2q_gate_combos(self):
        """Test the ctrl/tgt specific behavior for CX/ECR and default for others (like CZ).

        There are specific sequence for control/target/control+target spectator qubits for
        CX and ECR. Other gates do not get special sequences.

                     ┌────┐ ░      ┌──────┐
            q_0 -> 0 ┤ √X ├─░──────┤0     ├──────────
                     ├────┤ ░      │   ?  │
            q_1 -> 1 ┤ √X ├─░──────┤1     ├──────────
                     ├────┤ ░      └──────┘
            q_2 -> 2 ┤ √X ├─░── test this sequence ──
                     ├────┤ ░      ┌──────┐
            q_3 -> 3 ┤ √X ├─░──────┤0     ├──────────
                     ├────┤ ░      │   ?  │
            q_4 -> 4 ┤ √X ├─░──────┤1     ├──────────
                     └────┘ ░      └──────┘

        """
        target = get_toy_target(num_qubits=5, t_cx=self.t_cx)
        pm = _get_schedule_pm(target, list(range(5)))
        pm.append(ContextAwareDynamicalDecoupling(target))

        gates = ["cx", "ecr", "cz"]
        for top in gates:
            for bottom in gates:
                if top in ["cx", "ecr"]:
                    if bottom in ["cx", "ecr"]:
                        # squeezed between control and target
                        order = 2
                    else:
                        # other gate (CZ) has no specific behavior, use target-specific sequence
                        order = 0
                else:  # top gate has no specific behavior
                    if bottom in ["cx", "ecr"]:
                        # use ctrl-specific sequence
                        order = 1
                    else:  # both gates do not have special behavior, use 0
                        order = 0

                # construct circuit with specified gates
                circuit = QuantumCircuit(5)
                circuit.sx(circuit.qubits)
                circuit.barrier()
                getattr(circuit, top)(0, 1)
                getattr(circuit, bottom)(3, 4)
                circuit = pm.run(circuit)

                # compute reference
                ref = QuantumCircuit(5)
                ref.sx(ref.qubits)
                ref.barrier()
                getattr(ref, top)(0, 1)
                apply_delay_sequence(ref, 2, self.t_cx, target.durations(), order)
                getattr(ref, bottom)(3, 4)

                with self.subTest(top=top, bottom=bottom):
                    self.assertEqual(circuit, ref)

    def test_threshold_skipping(self):
        """Test skipping of delays that are below the threshold."""
        skip_threshold = 0.6  # arbitrary threshold below 1

        target = get_toy_target(num_qubits=5, t_x=self.t_x)
        pm = _get_schedule_pm(target, [0])
        pm.append(ContextAwareDynamicalDecoupling(target, skip_dd_threshold=skip_threshold))

        epsilon = 0.01  # ask a mathematician for a formal definition of epsilon
        dd_duration = 2 * self.t_x
        for exceed_threshold in [True, False]:
            # go above threshold once and below once
            if exceed_threshold:
                delay = math.floor(dd_duration / (skip_threshold + epsilon))
            else:
                delay = math.ceil(dd_duration / (skip_threshold - epsilon))

            circuit = QuantumCircuit(1)
            circuit.sx(0)
            circuit.delay(delay, 0)
            circuit.sx(0)

            circuit = pm.run(circuit)

            with self.subTest(exceed_threshold=exceed_threshold):
                expected_x = 0 if exceed_threshold else 2
                self.assertEqual(circuit.count_ops().get("x", 0), expected_x)

    @data(4, 20)  # X gate length is integer multiple of these values
    def test_pulse_alignment(self, alignment):
        """Test setting the pulse alignment.

        Pulses should only start at integer multiples of the allowed pulse alignment.
        """
        target = get_toy_target(num_qubits=5)
        pm = _get_schedule_pm(target, [0])
        pm.append(
            ContextAwareDynamicalDecoupling(
                target, pulse_alignment=alignment, skip_reset_qubits=False
            )
        )

        circuit = QuantumCircuit(1)
        circuit.delay(100, 0)
        circuit = pm.run(circuit)

        x_times = [
            time for gate, time in pm.property_set["node_start_time"].items() if gate.op.name == "x"
        ]

        self.assertTrue((np.asarray(x_times) % alignment == 0).all())

    def test_invalid_pulse_alignment(self):
        """Test an error is raised if the X gate length is not compatible with the pulse alignment."""
        target = get_toy_target(num_qubits=5, t_x=self.t_x)

        pm = _get_schedule_pm(target, [0])
        pm.append(
            ContextAwareDynamicalDecoupling(
                target, pulse_alignment=self.t_x + 1, skip_reset_qubits=False
            )
        )

        circuit = QuantumCircuit(1)
        circuit.delay(100, 0)

        with self.assertRaises(TranspilerError):
            _ = pm.run(circuit)

    def test_orthogonal_sequences(self):
        """Check orthogonality of sequences up to the supported order."""
        target = get_toy_target(5)
        cadd = ContextAwareDynamicalDecoupling(target)

        # get the DD sequences and check for the smallest time frame
        dd_sequences = [np.asarray(cadd.get_orthogonal_sequence(order)[0]) for order in range(6)]
        # we must exclude 0, which is sometimes used as padding in the end
        smallest_unit = np.min([np.min(row[np.where(row > 0)]) for row in dd_sequences])

        # expand the DD times to vectors with an entry +1 if the qubit is in + state
        # and a -1 if the qubit is in - state
        sign_vectors = [expand_dd_sequence(sequence, smallest_unit) for sequence in dd_sequences]

        # check pairwise orthogonality
        for i, row_i in enumerate(sign_vectors):
            for j, row_j in enumerate(sign_vectors[i + 1 :]):
                with self.subTest(i=i, j=j):
                    self.assertAlmostEqual(np.dot(row_i, row_j), 0)

    def test_exceed_default_highest_order(self):
        """Check an error is raised if we query a sequence with order too high."""
        order_threshold = 7  # pick an order above the default max order (5)
        kranka = Target(num_qubits=order_threshold, dt=self.dt)  # a target with scary connectivity
        coupling_web = itertools.combinations(list(range(order_threshold)), 2)
        cx_props = {
            connection: InstructionProperties(duration=self.t_cx * self.dt, error=1e-2)
            for connection in coupling_web
        }
        x_props = {
            (i,): InstructionProperties(duration=self.t_x * self.dt, error=1e-4)
            for i in range(kranka.num_qubits)
        }
        kranka.add_instruction(CXGate(), cx_props)
        kranka.add_instruction(XGate(), x_props)
        kranka.add_instruction(Delay(1), x_props)

        circuit = QuantumCircuit(order_threshold)
        circuit.x(circuit.qubits)
        circuit.barrier()
        circuit.delay(1000)  # note this should be long enough to fit the DD sequences

        pm = _get_schedule_pm(kranka, list(range(kranka.num_qubits)))
        pm.append(ContextAwareDynamicalDecoupling(kranka, skip_reset_qubits=False))

        circuit = pm.run(circuit)

        # number of flips in the first 7 Walsh-Hadamard sequences
        num_dd_x = 2 + 2 + 4 + 4 + 6 + 6 + 8
        num_x = circuit.num_qubits  # X layer
        self.assertEqual(num_x + num_dd_x, circuit.count_ops().get("x", 0))

    def test_collecting_pyramid(self):
        """Test collecting the largest adjacent blocks for a pyramid structure.

        The delays form the following circuit (just on more qubits in the test):

            q0: ------████-
            q1: ----██████-
            q2: -█████████-
            q3: ----██████-
            q4: ------████-

        The logs contain the multi-delay instructions, which we check here.
        """
        num_qubits = 9
        circuit = QuantumCircuit(num_qubits)
        center = num_qubits // 2
        circuit.sx(center)
        for i in range(num_qubits // 2):
            circuit.cx(center - i, center - (i + 1))
            circuit.cx(center + i, center + i + 1)

        target = get_toy_target(num_qubits=num_qubits, t_sx=self.t_sx, t_cx=self.t_cx)
        pm = _get_schedule_pm(target, list(range(num_qubits)))
        pm.append(ContextAwareDynamicalDecoupling(target))

        first_delay = [(1010, 2010, (center - 1,))]
        pyramid_delays = [
            (2010, 3010, (center,)),
            (3010, 4010, tuple(range(center - 1, center + 2))),
            (4010, 5010, tuple(range(center - 2, center + 3))),
        ]
        expected_blocks = [first_delay, pyramid_delays]

        with self.assertMultiDelayInserted(expected_blocks):
            _ = pm.run(circuit)

    def test_no_scheduling(self):
        """Test a sensible error is raised if no scheduling is performed prior to CA-DD."""
        circuit = QuantumCircuit(1)
        pass_ = ContextAwareDynamicalDecoupling(get_toy_target(1))

        with self.assertRaisesRegex(TranspilerError, "scheduling passes"):
            _ = pass_(circuit)

    def test_collecting_diamond_with_initial(self):
        """Test collecting delays in form of a diamond, including initial delays.

        This tests delay blocks are correctly merged when they get smaller again.

            q0: --████--
            q1: -██████-
            q2: --████--

        """
        circuit = QuantumCircuit(8)
        circuit.cx(2, 3)
        circuit.cx(1, 2)
        circuit.cx(0, 1)
        circuit.cx(0, 1)
        circuit.cx(1, 2)

        circuit.cx(4, 5)
        circuit.cx(5, 6)
        circuit.cx(6, 7)
        circuit.cx(6, 7)
        circuit.cx(5, 6)

        target = get_toy_target(num_qubits=circuit.num_qubits, t_sx=self.t_sx, t_cx=self.t_cx)
        pm = PassManager(
            [
                ASAPScheduleAnalysis(target.durations(), target),
                ContextAwareDynamicalDecoupling(target, skip_reset_qubits=False),
            ]
        )
        top_left = [(0, 1000, (0, 1)), (1000, 2000, (0,))]
        bottom_left = [
            (0, 1000, (6, 7)),
            (1000, 2000, (7,)),
        ]
        diamond = [(1000, 2000, (3, 4)), (2000, 4000, (2, 3, 4, 5)), (4000, 5000, (3, 4))]
        top_right = [(4000, 5000, (0,))]
        bottom_right = [(4000, 5000, (7,))]
        expected_blocks = [top_left, top_right, bottom_left, bottom_right, diamond]

        with self.assertMultiDelayInserted(expected_blocks):
            _ = pm.run(circuit)

    def test_collecting_donut(self):
        """Test collecting delays in a donut shape."""
        circuit = QuantumCircuit(3)
        circuit.delay(2 * self.t_cx, 0)
        circuit.delay(self.t_cx, 1)
        circuit.delay(2 * self.t_cx, 2)

        circuit.x(1)
        circuit.delay(self.t_cx - self.t_x, 1)

        target = get_toy_target(num_qubits=3, t_cx=self.t_cx, t_x=self.t_x)
        pm = PassManager(
            [
                ASAPScheduleAnalysis(target.durations(), target=target),
                ContextAwareDynamicalDecoupling(target, skip_reset_qubits=False),
            ]
        )

        expected_block = [
            (0, 1000, (0, 1, 2)),
            (1000, 1020, (0,)),
            (1000, 1020, (2,)),
            (1020, 2000, (0, 1, 2)),
        ]

        with self.assertMultiDelayInserted([expected_block]):
            _ = pm.run(circuit)

    def test_skip_after_reset(self):
        """Test delays after a reset are ignored.

        We're testing on a circuit looking like

            q0: -███-X-███-X-|0>-███-X-███
                  ^               ^
         ignored (initial)   ignored (after reset)

        """
        dt = 100
        circuit = QuantumCircuit(1)
        circuit.delay(dt, 0)
        circuit.x(0)
        circuit.delay(dt, 0)
        circuit.x(0)
        circuit.reset(0)
        circuit.delay(dt, 0)
        circuit.x(0)
        circuit.delay(dt, 0)

        target = get_toy_target(1)
        schedule_pm = _get_schedule_pm(target, list(range(circuit.num_qubits)))
        dd = PassManager(
            [
                ContextAwareDynamicalDecoupling(target, skip_reset_qubits=True),
            ]
        )
        pm = schedule_pm + dd
        circuit = pm.run(circuit)

        reference = QuantumCircuit(1)
        durations = target.durations()
        reference.delay(dt, 0)  # initial is ignored
        reference.x(0)
        apply_delay_sequence(reference, 0, dt, durations, order=0)
        reference.x(0)
        reference.reset(0)
        reference.delay(dt, 0)  # after reset is ignored
        reference.x(0)
        apply_delay_sequence(reference, 0, dt, durations, order=0)

        self.assertEqual(reference, circuit)

    def test_127q(self):
        """Test running on a larger backend with 100+ qubits.

        Since this is a bipartite graph, the coloring problem can be solved optimally, which
        in this case means the maximum order is 1.
        """
        backend = Mock127q()
        target = backend.target

        snake = (
            list(range(0, 14))[::-1]
            + [14]
            + list(range(18, 33))
            + [36]
            + list(range(37, 52))[::-1]
            + [52]
            + list(range(56, 71))
            + [74]
            + list(range(75, 90))[::-1]
            + [90]
            + list(range(94, 109))
            + [112]
            + list(range(113, 127))[::-1]
        )

        # build a circuit with staircase ECR gates in native directions, then we'll get
        # len(snake) - 2 concurrent delay blocks, which we can test for
        circuit = QuantumCircuit(backend.num_qubits)
        circuit.sx(snake)
        for ctrl, tgt in zip(snake[:-1], snake[1:]):
            if target.instruction_supported("ecr", (ctrl, tgt)):
                circuit.ecr(ctrl, tgt)
            else:
                circuit.ecr(tgt, ctrl)

        pre = generate_preset_pass_manager(
            optimization_level=0, target=target, initial_layout=list(range(127))
        )
        dd = PassManager(
            [
                ALAPScheduleAnalysis(target.durations(), target=target),
                ContextAwareDynamicalDecoupling(target, min_duration=0),
            ]
        )
        circuit = dd.run(pre.run(circuit))

        # compute the number of expected X gates via the number of DD sequences, which is in
        # the upper triangular part of the circuit, and each sequence has 2 X gates
        num_dd_sequences = int((len(snake) - 2) * (len(snake) - 1) / 2)
        num_x = 2 * num_dd_sequences

        # this can fail if the coloring is not solved optimally (i.e. we have colors >= 2)
        # or if there are more than len(snake) - 2 layers of multi delay operations
        self.assertEqual(num_x, circuit.count_ops().get("x", 0))

    def test_min_duration(self):
        """Test cutting off short peaks below the joinable duration.

        delay is too short
                   v
          q0: -----█------      q0: ------------
          q1: -██████████-  ->  q1: -██████████-
          q2: -----██████-      q2: -----██████-

        """
        circuit = QuantumCircuit(3)
        _ = [circuit.x([0, 2]) for _ in range(4)]
        circuit.delay(self.t_x, 0)  # this one should be ignored
        circuit.delay(10 * self.t_x, 1)
        circuit.delay(6 * self.t_x, 2)
        _ = [circuit.x(0) for _ in range(5)]

        num_x = circuit.count_ops()["x"]  # number of original X gates
        num_x += 2 + 2 + 2  # we expect 3 DD sequences -- without the cutoff, short delay

        target = get_toy_target(num_qubits=3, t_x=self.t_x)
        pm = PassManager(
            [
                ASAPScheduleAnalysis(target.durations(), target=target),
                ContextAwareDynamicalDecoupling(
                    target,
                    skip_reset_qubits=False,
                    min_duration=self.t_x + 1,
                ),
            ]
        )
        circuit = pm.run(circuit)

        self.assertEqual(circuit.count_ops().get("x", 0), num_x)

    @data(
        (rx.ColoringStrategy.Degree, 8),
        (rx.ColoringStrategy.IndependentSet, 10),
        (rx.ColoringStrategy.Saturation, 8),
    )
    @unpack
    def test_coloring_strategy(self, strategy, num_x):
        """Test setting different coloring strategies.

        Not all perform the same, e.g. IndependentSet is suboptimal on this qubit line
        with a boundary condition of the first idle wire being a target spectator.
        """
        circuit = QuantumCircuit(6)
        circuit.cx(0, 1)

        target = get_toy_target(circuit.num_qubits)
        pm = _get_schedule_pm(target, list(range(circuit.num_qubits)))
        pm.append(
            ContextAwareDynamicalDecoupling(
                target, skip_reset_qubits=False, coloring_strategy=strategy
            )
        )

        circuit = pm.run(circuit)

        self.assertEqual(num_x, circuit.count_ops().get("x", 0))


def _get_schedule_pm(target, initial_layout):
    durations = target.durations()

    schedule_pm = PassManager(
        [
            SetLayout(initial_layout),
            FullAncillaAllocation(target),
            EnlargeWithAncilla(),
            ApplyLayout(),
            ALAPScheduleAnalysis(durations, target),
        ]
    )

    return schedule_pm


def apply_delay_sequence(circuit, qubit, timespan, durations, order):
    """Tool to apply a delay sequence to a circuit."""
    if order == 0:
        dt = (timespan - 2 * durations.get("x", qubit)) / 2
        for _ in range(2):
            circuit.delay(dt, qubit)
            circuit.x(qubit)
    elif order == 1:
        reduced_timespan = timespan - 2 * durations.get("x", qubit)
        circuit.delay(reduced_timespan / 4, qubit)
        circuit.x(qubit)
        circuit.delay(reduced_timespan / 2, qubit)
        circuit.x(qubit)
        circuit.delay(reduced_timespan / 4, qubit)
    elif order == 2:
        dt = (timespan - 4 * durations.get("x", qubit)) / 4
        for _ in range(4):
            circuit.delay(dt, qubit)
            circuit.x(qubit)

    return circuit


def expand_dd_sequence(sequence, unit):
    """Expand durations in terms of the unit.

    E.g. in the unit is 1/8 and sequence is [1/4, 1/2, 1/4] expand to [++----++].
    """
    signs = []
    sign = 1

    for timespan in sequence:
        num_items = int(timespan / unit)
        signs += num_items * [sign]
        sign *= -1

    return signs


def get_toy_target(num_qubits, dt=1e-9, t_cx=1e3, t_sx=10, t_x=20):
    """Get a toy target."""

    # set up an idealistic target to test context-aware DD in a clean setting
    # (if I don't also add realistic settings I should be scolded)
    target = Target(num_qubits=num_qubits, dt=dt)
    # bidirectional linear next neighbor
    linear_topo = [(i, i + 1) for i in range(num_qubits - 1)]
    linear_topo += [tuple(reversed(connection)) for connection in linear_topo]
    # CX, SX and X gate durations (somewhat sensible durations and errors chosen)
    cx_props = {
        connection: InstructionProperties(duration=t_cx * dt, error=1e-2)
        for connection in linear_topo
    }
    sx_props = {
        (i,): InstructionProperties(duration=t_sx * dt, error=1e-4) for i in range(num_qubits)
    }
    x_props = {
        (i,): InstructionProperties(duration=t_x * dt, error=1e-4) for i in range(num_qubits)
    }
    target.add_instruction(CXGate(), cx_props)
    target.add_instruction(ECRGate(), cx_props)  # re-use CX props for ECR
    target.add_instruction(CZGate(), cx_props)  # re-use CX props for CZ
    target.add_instruction(SXGate(), sx_props)
    target.add_instruction(XGate(), x_props)
    target.add_instruction(Delay(1), sx_props)  # support delays, duration does not matter here
    target.add_instruction(Reset(), x_props)  # support resets, duration does not matter here

    return target


def get_varying_target(
    num_qubits, dt=1e-9, t_cx_range=(5e2, 5e3), t_sx_range=(10, 80), t_x_range=(20, 160)
):
    """Get a toy target."""

    # set up an idealistic target to test context-aware DD in a clean setting
    # (if I don't also add realistic settings I should be scolded)
    target = Target(num_qubits=num_qubits, dt=dt)
    # bidirectional linear next neighbor
    linear_topo = [(i, i + 1) for i in range(num_qubits - 1)]
    linear_topo += [tuple(reversed(connection)) for connection in linear_topo]
    # CX, SX and X gate durations (somewhat sensible durations and errors chosen)
    cx_props = {
        connection: InstructionProperties(duration=dt * np.random.randint(*t_cx_range), error=1e-2)
        for connection in linear_topo
    }
    sx_props = {
        (i,): InstructionProperties(duration=dt * np.random.randint(*t_sx_range), error=1e-4)
        for i in range(num_qubits)
    }
    x_props = {
        (i,): InstructionProperties(duration=dt * np.random.randint(*t_x_range), error=1e-4)
        for i in range(num_qubits)
    }
    target.add_instruction(CXGate(), cx_props)
    target.add_instruction(ECRGate(), cx_props)  # re-use CX props for ECR
    target.add_instruction(CZGate(), cx_props)  # re-use CX props for CZ
    target.add_instruction(SXGate(), sx_props)
    target.add_instruction(XGate(), x_props)
    target.add_instruction(Delay(1), sx_props)  # support delays, duration does not matter here
    target.add_instruction(Reset(), x_props)  # support resets, duration does not matter here

    # get the maximal time difference of the 2q gates, this is used for verification
    cx_durations = [prop.duration for prop in cx_props.values()]
    max_cx_diff = max(cx_durations) - min(cx_durations)

    return target, max_cx_diff



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Generic style visualization library.
"""

import json
import os
from typing import Any, Union
from warnings import warn
from pathlib import Path

from qiskit import user_config

from .exceptions import VisualizationError


class StyleDict(dict):
    """
    Attributes:
        VALID_FIELDS (set): Set of valid field inputs to a function that supports a style parameter
        ABBREVIATIONS (dict): Mapping of abbreviation:field for abbreviated inputs to VALID_FIELDS
            (must exist in VALID FIELDS)
        NESTED_ATTRS (set): Set of fields that are dictionaries, and need to be updated with .update
    """

    VALID_FIELDS = set()
    ABBREVIATIONS = {}
    NESTED_ATTRS = set()

    def __setitem__(self, key: Any, value: Any) -> None:
        # allow using field ABBREVIATIONS
        if key in self.ABBREVIATIONS:
            key = self.ABBREVIATIONS[key]

        if key not in self.VALID_FIELDS:
            warn(
                f"style option ({key}) is not supported",
                UserWarning,
                2,
            )
        return super().__setitem__(key, value)

    def __getitem__(self, key: Any) -> Any:
        # allow using field ABBREVIATIONS
        if key in self.ABBREVIATIONS:
            key = self.ABBREVIATIONS[key]

        return super().__getitem__(key)

    def update(self, other):
        # the attributes "displaycolor" and "displaytext" are dictionaries
        # themselves, therefore we need to propagate the update down to them
        for attr in self.NESTED_ATTRS.intersection(other.keys()):
            if attr in self.keys():
                self[attr].update(other[attr])
            else:
                self[attr] = other[attr]

        super().update((key, value) for key, value in other.items() if key not in self.NESTED_ATTRS)


class DefaultStyle:
    """
    Attributes:
        DEFAULT_STYLE_NAME (str): style name for the default style
        STYLE_PATH: file path where DEFAULT_STYLE_NAME.json is located
    """

    DEFAULT_STYLE_NAME = None
    DEFAULT_STYLE_PATH = None

    def __init__(self):
        raise NotImplementedError()


def load_style(
    style: Union[dict, str, None],
    style_dict: type[StyleDict],
    default_style: DefaultStyle,
    user_config_opt: str,
    user_config_path_opt: str,
    raise_error_if_not_found: bool = False,
) -> tuple[StyleDict, float]:
    """Utility function to load style from json files.

    Args:
        style: Depending on the type, this acts differently:

            * If a string, it can specify a supported style name (such
              as "iqp" or "clifford"). It can also specify the name of
              a custom color scheme stored as JSON file. This JSON file
              _must_ specify a complete set of colors.
            * If a dictionary, it may specify the style name via a
              ``{"name": "<desired style>"}`` entry. If this is not given,
              the default style will be used. The remaining entries in the
              dictionary can be used to override certain specs.
              E.g. ``{"name": "iqp", "ec": "#FF0000"}`` will use the ``"iqp"``
              color scheme but set the edgecolor to red.
        style_dict: The class used to define the options for loading styles
        default_style: DefaultStyle dictionary definition and documentation
        user_config_opt: User config field in the Qiskit User Configuration File
            used to define the style loaded
        user_config_path_opt: User config field in the Qiskit User Configuration File
            used to define the path to the style loaded
        raise_error_if_not_found: When True, load_style will throw a VisualizationError
            if the style parameter file is not found. When False, load_style will load
            the style passed in by the default_style parameter.


    Returns:
        A tuple containing the style as dictionary and the default font ratio.
    """

    default = default_style.DEFAULT_STYLE_NAME

    # if the style is not given, try to load the configured default (if set),
    # or use the default style
    config = user_config.get_config()
    if style is None:
        if config:
            style = config.get(user_config_opt, default)
        else:
            style = default

    # determine the style name which could also be inside a dictionary, like
    # style={"name": "clifford", <other settings...>}
    if isinstance(style, dict):
        style_name = style.get("name", default)
    elif isinstance(style, str):
        if style.endswith(".json"):
            style_name = style[:-5]
        else:
            style_name = style
    else:
        warn(
            f'Unsupported style parameter "{style}" of type {type(style)}. '
            "Will use the default style.",
            UserWarning,
            2,
        )
        style_name = default

    if style_name in [default]:
        current_style = default_style.style
    else:
        # Search for file in 'styles' dir, then config_path, and finally the current directory
        style_name = style_name + ".json"
        style_paths = []

        default_path = default_style.DEFAULT_STYLE_PATH / style_name
        style_paths.append(default_path)

        # check configured paths, if there are any
        if config:
            config_path = config.get(user_config_path_opt, "")
            if config_path:
                for path in config_path:
                    style_paths.append(Path(path) / style_name)

        # check current directory
        cwd_path = Path("") / style_name
        style_paths.append(cwd_path)

        for path in style_paths:
            # expand ~ to the user directory and check if the file exists
            exp_user = path.expanduser()
            if os.path.isfile(exp_user):
                try:
                    with open(exp_user) as infile:
                        json_style = json.load(infile)

                    current_style = style_dict(json_style)
                    break
                except json.JSONDecodeError as err:
                    warn(
                        f"Could not decode JSON in file '{path}': {str(err)}. "
                        "Will use default style.",
                        UserWarning,
                        2,
                    )
                    break
                except (OSError, FileNotFoundError):
                    warn(
                        f"Error loading JSON file '{path}'. Will use default style.",
                        UserWarning,
                        2,
                    )
                    break
        else:
            if raise_error_if_not_found:
                raise VisualizationError(f"Invalid style {style_name}")

            warn(
                f"Style JSON file '{style_name}' not found in any of these locations: "
                f"{', '.join(map(str, style_paths))}. "
                "Will use default style.",
                UserWarning,
                2,
            )
            current_style = default_style.style

    # if the style is a dictionary, update the defaults with the new values
    # this _needs_ to happen after loading by name to cover cases like
    #   style = {"name": "bw", "edgecolor": "#FF0000"}
    if isinstance(style, dict):
        current_style.update(style)

    # this is the default font ratio
    # if the font- or subfont-sizes are changed, the new size is based on this ratio
    def_font_ratio = 13 / 8

    return current_style, def_font_ratio
